# This file is part of allegedb, an object relational mapper for graphs.
# Copyright (c) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Wrapper to run SQL queries in a lightly abstracted way, such that
code that's more to do with the queries than with the data per se
doesn't pollute the other files so much.

"""
from threading import Thread, Lock
from time import monotonic
from typing import Tuple, Any, Iterator, Hashable
from queue import Queue
import os
from collections.abc import MutableMapping
import sqlite3

from . import wrap

wrappath = os.path.dirname(wrap.__file__)
alchemyIntegError = None
alchemyOperationalError = None
try:
	import sqlalchemy.exc

	IntegrityError = (sqlalchemy.exc.IntegrityError, sqlite3.IntegrityError)
	OperationalError = (sqlalchemy.exc.OperationalError,
						sqlite3.OperationalError)
except ImportError:
	IntegrityError = sqlite3.IntegrityError
	OperationalError = sqlite3.OperationalError

NodeRowType = Tuple[Hashable, Hashable, str, int, int, bool]
EdgeRowType = Tuple[Hashable, Hashable, Hashable, int, str, int, int, bool]
GraphValRowType = Tuple[Hashable, Hashable, str, int, int, Any]
NodeValRowType = Tuple[Hashable, Hashable, Hashable, str, int, int, Any]
EdgeValRowType = Tuple[Hashable, Hashable, Hashable, int, str, int, int, Any]


class TimeError(ValueError):
	"""Exception class for problems with the time model"""


class GlobalKeyValueStore(MutableMapping):
	"""A dict-like object that keeps its contents in a table.

    Mostly this is for holding the current branch and revision.

    """

	def __init__(self, qe):
		self.qe = qe
		self._cache = dict(qe.global_items())

	def __iter__(self):
		yield from self._cache

	def __len__(self):
		return len(self._cache)

	def __getitem__(self, k):
		return self._cache[k]

	def __setitem__(self, k, v):
		self.qe.global_set(k, v)
		self._cache[k] = v

	def __delitem__(self, k):
		del self._cache[k]
		self.qe.global_del(k)


class ConnectionHolder:

	def __init__(self, dbstring, connect_args, alchemy, inq, outq, fn, tables):
		self.lock = Lock()
		self.existence_lock = Lock()
		self.existence_lock.acquire()
		self._dbstring = dbstring
		self._connect_args = connect_args
		self._alchemy = alchemy
		self._fn = fn
		self.inq = inq
		self.outq = outq
		self.tables = tables

	def commit(self):
		if hasattr(self, 'transaction') and self.transaction.is_active:
			self.transaction.commit()
		elif hasattr(self, 'connection'):
			self.connection.commit()

	def init_table(self, tbl):
		return self.sql('create_{}'.format(tbl))

	def sql(self, stringname, *args, **kwargs):
		"""Wrapper for the various prewritten or compiled SQL calls.

        First argument is the name of the query, either a key in
        ``sqlite.json`` or a method name in
        ``allegedb.alchemy.Alchemist``. The rest of the arguments are
        parameters to the query.

        """
		if hasattr(self, 'alchemist'):
			return getattr(self.alchemist, stringname)(*args, **kwargs)
		else:
			s = self.strings[stringname]
			return self.connection.cursor().execute(
				s.format(**kwargs) if kwargs else s, args)

	def sqlmany(self, stringname, *args):
		"""Wrapper for executing many SQL calls on my connection.

        First arg is the name of a query, either a key in the
        precompiled JSON or a method name in
        ``allegedb.alchemy.Alchemist``. Remaining arguments should be
        tuples of argument sequences to be passed to the query.

        """
		if hasattr(self, 'alchemist'):
			return getattr(self.alchemist.many, stringname)(*args)
		s = self.strings[stringname]
		return self.connection.cursor().executemany(s, args)

	def run(self):
		alchemy = self._alchemy
		dbstring = self._dbstring
		connect_args = self._connect_args

		def lite_init(dbstring):
			from json import load
			with open(self._fn, "rb") as strf:
				self.strings = load(strf)
			assert not isinstance(dbstring, sqlite3.Connection)
			if dbstring.startswith('sqlite:///'):
				dbstring = dbstring[10:]
			self.connection = sqlite3.connect(dbstring)

		if alchemy:
			try:
				from sqlalchemy import create_engine
				from sqlalchemy.engine.base import Engine
				from sqlalchemy.exc import ArgumentError
				from sqlalchemy.pool import NullPool
				from .alchemy import Alchemist
				if isinstance(dbstring, Engine):
					self.engine = dbstring
				else:
					try:
						self.engine = create_engine(dbstring,
													connect_args=connect_args,
													poolclass=NullPool)
					except ArgumentError:
						self.engine = create_engine('sqlite:///' + dbstring,
													connect_args=connect_args,
													poolclass=NullPool)
				self.alchemist = Alchemist(self.engine)
				self.transaction = self.alchemist.conn.begin()
			except ImportError:
				self._alchemy = False
				lite_init(dbstring)
		else:
			lite_init(dbstring)
		while True:
			inst = self.inq.get()
			if inst == 'shutdown':
				self.commit()
				if hasattr(self, 'alchemist'):
					self.alchemist.conn.close()
					self.engine.dispose()
				else:
					self.connection.close()
				self.existence_lock.release()
				return
			if inst == 'commit':
				self.commit()
				continue
			if inst == 'initdb':
				self.outq.put(self.initdb())
				continue
			silent = False
			if inst[0] == 'silent':
				inst = inst[1:]
				silent = True
			if inst[0] == 'one':
				try:
					res = self.sql(inst[1], *inst[2], **inst[3])
					if not silent:
						if hasattr(res, 'returns_rows'):
							if res.returns_rows:
								o = list(res)
								self.outq.put(o)
							else:
								self.outq.put(None)
						else:
							o = list(res)
							self.outq.put(o)
				except Exception as ex:
					if not silent:
						self.outq.put(ex)
			elif inst[0] != 'many':
				raise ValueError(f"Invalid instruction: {inst[0]}")
			else:
				try:
					res = self.sqlmany(inst[1], *inst[2])
					if not silent:
						if hasattr(res, 'returns_rows'):
							if res.returns_rows:
								self.outq.put(list(res))
							else:
								self.outq.put(None)
						else:
							rez = list(res.fetchall())
							self.outq.put(rez or None)
				except Exception as ex:
					if not silent:
						self.outq.put(ex)

	def initdb(self):
		"""Create tables and indices as needed."""
		for table in ('branches', 'turns', 'graphs', 'graph_val', 'nodes',
						'node_val', 'edges', 'edge_val', 'plans', 'plan_ticks',
						'keyframes', 'global'):
			try:
				ret = self.init_table(table)
			except OperationalError:
				pass
			except Exception as ex:
				return ex
		self.commit()


class QueryEngine(object):
	"""Wrapper around either a DBAPI2.0 connection or an
    Alchemist. Provides methods to run queries using either.

    """
	flush_edges_t = 0
	holder_cls = ConnectionHolder
	tables = ('global', 'branches', 'turns', 'graphs', 'keyframes',
				'graph_val', 'nodes', 'node_val', 'edges', 'edge_val', 'plans',
				'plan_ticks', 'universals')

	def __init__(self,
					dbstring,
					connect_args,
					alchemy,
					strings_filename: str = None,
					pack=None,
					unpack=None):
		"""If ``alchemy`` is True and ``dbstring`` is a legit database URI,
        instantiate an Alchemist and start a transaction with
        it. Otherwise use sqlite3.

        You may pass an already created sqlalchemy :class:`Engine`
        object in place of ``dbstring`` if you wish. I'll still create
        my own transaction though.

        """
		dbstring = dbstring or 'sqlite:///:memory:'
		self._inq = Queue()
		self._outq = Queue()
		if strings_filename is None:
			strings_filename = os.path.join(
				os.path.dirname(os.path.abspath(__file__)), "sqlite.json")
		if not os.path.exists(strings_filename) or os.path.isdir(
			strings_filename):
			raise FileNotFoundError("No SQL in JSON found at " +
									strings_filename)
		self._holder = self.holder_cls(dbstring, connect_args, alchemy,
										self._inq, self._outq,
										strings_filename, self.tables)

		if unpack is None:
			from ast import literal_eval as unpack
		self.pack = pack or repr
		self.unpack = unpack
		self._branches = {}
		self._nodevals2set = []
		self._edgevals2set = []
		self._graphvals2set = []
		self._nodes2set = []
		self._edges2set = []
		self._btts = set()
		self._t = Thread(target=self._holder.run, daemon=True)
		self._t.start()

	def sql(self, string, *args, **kwargs):
		__doc__ = ConnectionHolder.sql.__doc__
		with self._holder.lock:
			self._inq.put(('one', string, args, kwargs))
			ret = self._outq.get()
		if isinstance(ret, Exception):
			raise ret
		return ret

	def sqlmany(self, string, *args):
		__doc__ = ConnectionHolder.sqlmany.__doc__
		with self._holder.lock:
			self._inq.put(('many', string, args))
			ret = self._outq.get()
		if isinstance(ret, Exception):
			raise ret
		return ret

	def have_graph(self, graph):
		"""Return whether I have a graph by this name."""
		graph = self.pack(graph)
		return bool(self.sql('graphs_named', graph)[0][0])

	def new_graph(self, graph, typ):
		"""Declare a new graph by this name of this type."""
		graph = self.pack(graph)
		return self.sql('graphs_insert', graph, typ)

	def keyframes_insert(self, graph, branch, turn, tick, nodes, edges,
							graph_val):
		graph, nodes, edges, graph_val = map(
			self.pack, (graph, nodes, edges, graph_val))
		return self.sql('keyframes_insert', graph, branch, turn, tick, nodes,
						edges, graph_val)

	def keyframes_insert_many(self, many):
		pack = self.pack
		return self.sqlmany(
			'keyframes_insert',
			*[(pack(graph), branch, turn, tick, pack(nodes), pack(edges),
				pack(graph_val)) for (graph, branch, turn, tick, nodes, edges,
										graph_val) in many])

	def keyframes_dump(self):
		unpack = self.unpack
		for (graph, branch, turn, tick, nodes, edges,
				graph_val) in self.sql('keyframes_dump'):
			yield unpack(graph), branch, turn, tick, unpack(nodes), unpack(
				edges), unpack(graph_val)

	def keyframes_list(self):
		unpack = self.unpack
		for (graph, branch, turn, tick) in self.sql('keyframes_list'):
			yield unpack(graph), branch, turn, tick

	def get_keyframe(self, graph, branch, turn, tick):
		unpack = self.unpack
		stuff = self.sql('get_keyframe', self.pack(graph), branch, turn, tick)
		if not stuff:
			return
		nodes, edges, graph_val = stuff[0]
		return unpack(nodes), unpack(edges), unpack(graph_val)

	def del_graph(self, graph):
		"""Delete all records to do with the graph"""
		g = self.pack(graph)
		self.sql('del_edge_val_graph', g)
		self.sql('del_node_val_graph', g)
		self.sql('del_edge_val_graph', g)
		self.sql('del_edges_graph', g)
		self.sql('del_nodes_graph', g)
		self.sql('del_graph', g)

	def graph_type(self, graph):
		"""What type of graph is this?"""
		graph = self.pack(graph)
		return self.sql('graph_type', graph)[0][0]

	def have_branch(self, branch):
		"""Return whether the branch thus named exists in the database."""
		return bool(self.sql('ctbranch', branch)[0][0])

	def all_branches(self):
		"""Return all the branch data in tuples of (branch, parent,
        parent_turn).

        """
		return self.sql('branches_dump')

	def global_get(self, key):
		"""Return the value for the given key in the ``globals`` table."""
		key = self.pack(key)
		r = self.sql('global_get', key)[0]
		if r is None:
			raise KeyError("Not set")
		return self.unpack(r[0])

	def global_items(self):
		"""Iterate over (key, value) pairs in the ``globals`` table."""
		unpack = self.unpack
		dumped = self.sql('global_dump')
		for (k, v) in dumped:
			yield (unpack(k), unpack(v))

	def get_branch(self):
		v = self.sql('global_get', self.pack('branch'))[0]
		if v is None:
			return 'trunk'
		return self.unpack(v[0])

	def get_turn(self):
		v = self.sql('global_get', self.pack('turn'))[0]
		if v is None:
			return 0
		return self.unpack(v[0])

	def get_tick(self):
		v = self.sql('global_get', self.pack('tick'))[0]
		if v is None:
			return 0
		return self.unpack(v[0])

	def global_set(self, key, value):
		"""Set ``key`` to ``value`` globally (not at any particular branch or
        revision)

        """
		(key, value) = map(self.pack, (key, value))
		try:
			return self.sql('global_insert', key, value)
		except IntegrityError:
			return self.sql('global_update', value, key)

	def global_del(self, key):
		"""Delete the global record for the key."""
		key = self.pack(key)
		return self.sql('global_del', key)

	def new_branch(self, branch, parent, parent_turn, parent_tick):
		"""Declare that the ``branch`` is descended from ``parent`` at
        ``parent_turn``, ``parent_tick``

        """
		return self.sql('branches_insert', branch, parent, parent_turn,
						parent_tick, parent_turn, parent_tick)

	def update_branch(self, branch, parent, parent_turn, parent_tick, end_turn,
						end_tick):
		return self.sql('update_branches', parent, parent_turn, parent_tick,
						end_turn, end_tick, branch)

	def set_branch(self, branch, parent, parent_turn, parent_tick, end_turn,
					end_tick):
		try:
			self.sql('branches_insert', branch, parent, parent_turn,
						parent_tick, end_turn, end_tick)
		except IntegrityError:
			self.update_branch(branch, parent, parent_turn, parent_tick,
								end_turn, end_tick)

	def new_turn(self, branch, turn, end_tick=0, plan_end_tick=0):
		return self.sql('turns_insert', branch, turn, end_tick, plan_end_tick)

	def update_turn(self, branch, turn, end_tick, plan_end_tick):
		return self.sql('update_turns', end_tick, plan_end_tick, branch, turn)

	def set_turn(self, branch, turn, end_tick, plan_end_tick):
		try:
			return self.sql('turns_insert', branch, turn, end_tick,
							plan_end_tick)
		except IntegrityError:
			return self.sql('update_turns', end_tick, plan_end_tick, branch,
							turn)

	def set_turn_completed(self, branch, turn):
		try:
			return self.sql('turns_completed_insert', branch, turn)
		except IntegrityError:
			return self.sql('turns_completed_update', turn, branch)

	def turns_dump(self):
		return self.sql('turns_dump')

	def graph_val_dump(self) -> Iterator[GraphValRowType]:
		"""Yield the entire contents of the graph_val table."""
		self._flush_graph_val()
		unpack = self.unpack
		for (graph, key, branch, turn, tick,
				value) in self.sql('graph_val_dump'):
			yield (unpack(graph), unpack(key), branch, turn, tick,
					unpack(value))

	def load_graph_val(self,
						graph,
						branch,
						turn_from,
						tick_from,
						turn_to=None,
						tick_to=None) -> Iterator[GraphValRowType]:
		if (turn_to is None) ^ (tick_to is None):
			raise TypeError("I need both or neither of turn_to and tick_to")
		self._flush_graph_val()
		pack = self.pack
		unpack = self.unpack
		if turn_to is None:
			it = self.sql('load_graph_val_tick_to_end', pack(graph), branch,
							turn_from, turn_from, tick_from)
		else:
			it = self.sql('load_graph_val_tick_to_tick', pack(graph), branch,
							turn_from, turn_from, tick_from, turn_to, turn_to,
							tick_to)
		for (key, turn, tick, value) in it:
			yield graph, unpack(key), branch, turn, tick, unpack(value)

	def _flush_graph_val(self):
		"""Send all new and changed graph values to the database."""
		if not self._graphvals2set:
			return
		pack = self.pack
		self.sqlmany(
			'graph_val_insert',
			*((pack(graph), pack(key), branch, turn, tick, pack(value))
				for (graph, key, branch, turn, tick,
						value) in self._graphvals2set))
		self._graphvals2set = []

	def graph_val_set(self, graph, key, branch, turn, tick, value):
		if (branch, turn, tick) in self._btts:
			raise TimeError
		self._btts.add((branch, turn, tick))
		self._graphvals2set.append((graph, key, branch, turn, tick, value))

	def graph_val_del_time(self, branch, turn, tick):
		self._flush_graph_val()
		self.sql('graph_val_del_time', branch, turn, tick)
		self._btts.discard((branch, turn, tick))

	def graphs_types(self):
		for (graph, typ) in self.sql('graphs_types'):
			yield (self.unpack(graph), typ)

	def _flush_nodes(self):
		if not self._nodes2set:
			return
		pack = self.pack
		self.sqlmany(
			'nodes_insert',
			*((pack(graph), pack(node), branch, turn, tick, bool(extant))
				for (graph, node, branch, turn, tick,
						extant) in self._nodes2set))
		self._nodes2set = []

	def exist_node(self, graph, node, branch, turn, tick, extant):
		"""Declare that the node exists or doesn't.

        Inserts a new record or updates an old one, as needed.

        """
		if (branch, turn, tick) in self._btts:
			raise TimeError
		self._btts.add((branch, turn, tick))
		self._nodes2set.append((graph, node, branch, turn, tick, extant))

	def nodes_del_time(self, branch, turn, tick):
		self._flush_nodes()
		self.sql('nodes_del_time', branch, turn, tick)
		self._btts.discard((branch, turn, tick))

	def nodes_dump(self) -> Iterator[NodeRowType]:
		"""Dump the entire contents of the nodes table."""
		self._flush_nodes()
		unpack = self.unpack
		for (graph, node, branch, turn, tick,
				extant) in self.sql('nodes_dump'):
			yield (unpack(graph), unpack(node), branch, turn, tick,
					bool(extant))

	def load_nodes(self,
					graph,
					branch,
					turn_from,
					tick_from,
					turn_to=None,
					tick_to=None) -> Iterator[NodeRowType]:
		if (turn_to is None) ^ (tick_to is None):
			raise TypeError("I need both or neither of turn_to and tick_to")
		self._flush_nodes()
		pack = self.pack
		unpack = self.unpack
		if turn_to is None:
			it = self.sql('load_nodes_tick_to_end', pack(graph), branch,
							turn_from, turn_from, tick_from)
		else:
			it = self.sql('load_nodes_tick_to_tick', pack(graph), branch,
							turn_from, turn_from, tick_from, turn_to, turn_to,
							tick_to)
		for (node, turn, tick, extant) in it:
			yield graph, unpack(node), branch, turn, tick, extant

	def node_val_dump(self) -> Iterator[NodeValRowType]:
		"""Yield the entire contents of the node_val table."""
		self._flush_node_val()
		unpack = self.unpack
		for (graph, node, key, branch, turn, tick,
				value) in self.sql('node_val_dump'):
			yield (unpack(graph), unpack(node), unpack(key), branch, turn,
					tick, unpack(value))

	def load_node_val(self,
						graph,
						branch,
						turn_from,
						tick_from,
						turn_to=None,
						tick_to=None) -> Iterator[NodeValRowType]:
		if (turn_to is None) ^ (tick_to is None):
			raise TypeError("I need both or neither of turn_to and tick_to")
		self._flush_node_val()
		pack = self.pack
		unpack = self.unpack
		if turn_to is None:
			it = self.sql('load_node_val_tick_to_end', pack(graph), branch,
							turn_from, turn_from, tick_from)
		else:
			it = self.sql('load_node_val_tick_to_tick', pack(graph), branch,
							turn_from, turn_from, tick_from, turn_to, turn_to,
							tick_to)
		for (node, key, turn, tick, value) in it:
			yield graph, unpack(node), unpack(key), branch, turn, tick, unpack(
				value)

	def _flush_node_val(self):
		if not self._nodevals2set:
			return
		pack = self.pack
		self.sqlmany(
			'node_val_insert',
			*((pack(graph), pack(node), pack(key), branch, turn, tick,
				pack(value)) for (graph, node, key, branch, turn, tick,
									value) in self._nodevals2set))
		self._nodevals2set = []

	def node_val_set(self, graph, node, key, branch, turn, tick, value):
		"""Set a key-value pair on a node at a specific branch and revision"""
		if (branch, turn, tick) in self._btts:
			raise TimeError
		self._btts.add((branch, turn, tick))
		self._nodevals2set.append(
			(graph, node, key, branch, turn, tick, value))

	def node_val_del_time(self, branch, turn, tick):
		self._flush_node_val()
		self.sql('node_val_del_time', branch, turn, tick)
		self._btts.discard((branch, turn, tick))

	def edges_dump(self) -> Iterator[EdgeRowType]:
		"""Dump the entire contents of the edges table."""
		self._flush_edges()
		unpack = self.unpack
		for (graph, orig, dest, idx, branch, turn, tick,
				extant) in self.sql('edges_dump'):
			yield (unpack(graph), unpack(orig), unpack(dest), idx, branch,
					turn, tick, bool(extant))

	def load_edges(self,
					graph,
					branch,
					turn_from,
					tick_from,
					turn_to=None,
					tick_to=None) -> Iterator[EdgeRowType]:
		if (turn_to is None) ^ (tick_to is None):
			raise TypeError("I need both or neither of turn_to and tick_to")
		self._flush_edge_val()
		pack = self.pack
		unpack = self.unpack
		if turn_to is None:
			it = self.sql('load_edges_tick_to_end', pack(graph), branch,
							turn_from, turn_from, tick_from)
		else:
			it = self.sql('load_edges_tick_to_tick', pack(graph), branch,
							turn_from, turn_from, tick_from, turn_to, turn_to,
							tick_to)
		for (orig, dest, idx, turn, tick, extant) in it:
			yield graph, unpack(orig), unpack(
				dest), idx, branch, turn, tick, extant

	def _pack_edge2set(self, tup):
		graph, orig, dest, idx, branch, turn, tick, extant = tup
		pack = self.pack
		return pack(graph), pack(orig), pack(
			dest), idx, branch, turn, tick, extant

	def _flush_edges(self):
		start = monotonic()
		if not self._edges2set:
			return
		self.sqlmany('edges_insert', *map(self._pack_edge2set,
											self._edges2set))
		self._edges2set = []
		QueryEngine.flush_edges_t += monotonic() - start

	def exist_edge(self, graph, orig, dest, idx, branch, turn, tick, extant):
		"""Declare whether or not this edge exists."""
		if (branch, turn, tick) in self._btts:
			raise TimeError
		self._btts.add((branch, turn, tick))
		self._edges2set.append(
			(graph, orig, dest, idx, branch, turn, tick, extant))

	def edges_del_time(self, branch, turn, tick):
		self._flush_edges()
		self.sql('edges_del_time', branch, turn, tick)
		self._btts.discard((branch, turn, tick))

	def edge_val_dump(self) -> Iterator[EdgeValRowType]:
		"""Yield the entire contents of the edge_val table."""
		self._flush_edge_val()
		unpack = self.unpack
		for (graph, orig, dest, idx, key, branch, turn, tick,
				value) in self.sql('edge_val_dump'):
			yield (unpack(graph), unpack(orig), unpack(dest), idx, unpack(key),
					branch, turn, tick, unpack(value))

	def load_edge_val(self,
						graph,
						branch,
						turn_from,
						tick_from,
						turn_to=None,
						tick_to=None) -> Iterator[EdgeValRowType]:
		if (turn_to is None) ^ (tick_to is None):
			raise TypeError("I need both or neither of turn_to and tick_to")
		self._flush_edge_val()
		pack = self.pack
		unpack = self.unpack
		if turn_to is None:
			it = self.sql('load_edge_val_tick_to_end', pack(graph), branch,
							turn_from, turn_from, tick_from)
		else:
			it = self.sql('load_edge_val_tick_to_tick', pack(graph), branch,
							turn_from, turn_from, tick_from, turn_to, turn_to,
							tick_to)
		for (orig, dest, idx, key, turn, tick, value) in it:
			yield graph, unpack(orig), unpack(dest), idx, unpack(
				key), branch, turn, tick, unpack(value)

	def _pack_edgeval2set(self, tup):
		graph, orig, dest, idx, key, branch, turn, tick, value = tup
		pack = self.pack
		return pack(graph), pack(orig), pack(dest), idx, pack(
			key), branch, turn, tick, pack(value)

	def _flush_edge_val(self):
		if not self._edgevals2set:
			return
		self.sqlmany('edge_val_insert',
						*map(self._pack_edgeval2set, self._edgevals2set))
		self._edgevals2set = []

	def edge_val_set(self, graph, orig, dest, idx, key, branch, turn, tick,
						value):
		"""Set this key of this edge to this value."""
		if (branch, turn, tick) in self._btts:
			raise TimeError
		self._btts.add((branch, turn, tick))
		self._edgevals2set.append(
			(graph, orig, dest, idx, key, branch, turn, tick, value))

	def edge_val_del_time(self, branch, turn, tick):
		self._flush_edge_val()
		self.sql('edge_val_del_time', branch, turn, tick)
		self._btts.discard((branch, turn, tick))

	def plans_dump(self):
		return self.sql('plans_dump')

	def plans_insert(self, plan_id, branch, turn, tick):
		return self.sql('plans_insert', plan_id, branch, turn, tick)

	def plans_insert_many(self, many):
		return self.sqlmany('plans_insert', *many)

	def plan_ticks_insert(self, plan_id, turn, tick):
		return self.sql('plan_ticks_insert', plan_id, turn, tick)

	def plan_ticks_insert_many(self, many):
		return self.sqlmany('plan_ticks_insert', *many)

	def plan_ticks_dump(self):
		return self.sql('plan_ticks_dump')

	def flush(self):
		"""Put all pending changes into the SQL transaction."""
		self._flush_nodes()
		self._flush_edges()
		self._flush_graph_val()
		self._flush_node_val()
		self._flush_edge_val()

	def commit(self):
		"""Commit the transaction"""
		self.flush()
		self._inq.put('commit')

	def close(self):
		"""Commit the transaction, then close the connection"""
		self._inq.put('shutdown')
		self._holder.existence_lock.acquire()
		self._holder.existence_lock.release()
		self._t.join()

	def initdb(self):
		with self._holder.lock:
			self._inq.put('initdb')
			ret = self._outq.get()
			if isinstance(ret, Exception):
				raise ret
		self.globl = GlobalKeyValueStore(self)
		if 'branch' not in self.globl:
			self.globl['branch'] = 'trunk'
		if 'turn' not in self.globl:
			self.globl['turn'] = 0
		if 'tick' not in self.globl:
			self.globl['tick'] = 0

	def truncate_all(self):
		"""Delete all data from every table"""
		for table in self.tables:
			try:
				self.sql('truncate_' + table)
			except OperationalError:
				pass  # table wasn't created yet
		self.commit()



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
PHYSICAL_INITIAL_COPY = {
	'node_val': {
		'common0': {
			'rulebook': ('physical', 'common0')
		},
		'dorm1room3': {
			'rulebook': ('physical', 'dorm1room3')
		},
		'dorm1room4student1': {
			'rulebook': ('physical', 'dorm1room4student1'),
			'location': 'dorm1room4'
		},
		'dorm2room4student0': {
			'rulebook': ('physical', 'dorm2room4student0'),
			'location': 'dorm2room4'
		},
		'classroom': {
			'rulebook': ('physical', 'classroom')
		},
		'dorm0room5student1': {
			'rulebook': ('physical', 'dorm0room5student1'),
			'location': 'dorm0room5'
		},
		'dorm2room3student1': {
			'rulebook': ('physical', 'dorm2room3student1'),
			'location': 'dorm2room3'
		},
		'dorm0room3student0': {
			'rulebook': ('physical', 'dorm0room3student0'),
			'location': 'dorm0room3'
		},
		'dorm0room2student0': {
			'rulebook': ('physical', 'dorm0room2student0'),
			'location': 'dorm0room2'
		},
		'dorm2room1student0': {
			'rulebook': ('physical', 'dorm2room1student0'),
			'location': 'dorm2room1'
		},
		'dorm1room4student0': {
			'rulebook': ('physical', 'dorm1room4student0'),
			'location': 'dorm1room4'
		},
		'dorm2room0student1': {
			'rulebook': ('physical', 'dorm2room0student1'),
			'location': 'dorm2room0'
		},
		'dorm1room1': {
			'rulebook': ('physical', 'dorm1room1')
		},
		'dorm2room0': {
			'rulebook': ('physical', 'dorm2room0')
		},
		'dorm2room3student0': {
			'rulebook': ('physical', 'dorm2room3student0'),
			'location': 'dorm2room3'
		},
		'dorm0room0student0': {
			'rulebook': ('physical', 'dorm0room0student0'),
			'location': 'dorm0room0'
		},
		'dorm2room2student0': {
			'rulebook': ('physical', 'dorm2room2student0'),
			'location': 'dorm2room2'
		},
		'dorm0room3student1': {
			'rulebook': ('physical', 'dorm0room3student1'),
			'location': 'dorm0room3'
		},
		'dorm0room5': {
			'rulebook': ('physical', 'dorm0room5')
		},
		'dorm2room1': {
			'rulebook': ('physical', 'dorm2room1')
		},
		'dorm1room1student1': {
			'rulebook': ('physical', 'dorm1room1student1'),
			'location': 'dorm1room1'
		},
		'dorm0room4student0': {
			'rulebook': ('physical', 'dorm0room4student0'),
			'location': 'dorm0room4'
		},
		'dorm2room2student1': {
			'rulebook': ('physical', 'dorm2room2student1'),
			'location': 'dorm2room2'
		},
		'dorm1room2': {
			'rulebook': ('physical', 'dorm1room2')
		},
		'dorm0room2': {
			'rulebook': ('physical', 'dorm0room2')
		},
		'dorm2room5': {
			'rulebook': ('physical', 'dorm2room5')
		},
		'dorm1room3student1': {
			'rulebook': ('physical', 'dorm1room3student1'),
			'location': 'dorm1room3'
		},
		'dorm0room2student1': {
			'rulebook': ('physical', 'dorm0room2student1'),
			'location': 'dorm0room2'
		},
		'dorm1room5student0': {
			'rulebook': ('physical', 'dorm1room5student0'),
			'location': 'dorm1room5'
		},
		'dorm1room0student0': {
			'rulebook': ('physical', 'dorm1room0student0'),
			'location': 'dorm1room0'
		},
		'dorm2room5student0': {
			'rulebook': ('physical', 'dorm2room5student0'),
			'location': 'dorm2room5'
		},
		'dorm1room5student1': {
			'rulebook': ('physical', 'dorm1room5student1'),
			'location': 'dorm1room5'
		},
		'dorm1room4': {
			'rulebook': ('physical', 'dorm1room4')
		},
		'dorm0room1student1': {
			'rulebook': ('physical', 'dorm0room1student1'),
			'location': 'dorm0room1'
		},
		'dorm1room2student0': {
			'rulebook': ('physical', 'dorm1room2student0'),
			'location': 'dorm1room2'
		},
		'common2': {
			'rulebook': ('physical', 'common2')
		},
		'dorm2room1student1': {
			'rulebook': ('physical', 'dorm2room1student1'),
			'location': 'dorm2room1'
		},
		'dorm1room0student1': {
			'rulebook': ('physical', 'dorm1room0student1'),
			'location': 'dorm1room0'
		},
		'dorm1room2student1': {
			'rulebook': ('physical', 'dorm1room2student1'),
			'location': 'dorm1room2'
		},
		'dorm0room0student1': {
			'rulebook': ('physical', 'dorm0room0student1'),
			'location': 'dorm0room0'
		},
		'dorm0room1student0': {
			'rulebook': ('physical', 'dorm0room1student0'),
			'location': 'dorm0room1'
		},
		'dorm1room0': {
			'rulebook': ('physical', 'dorm1room0')
		},
		'dorm2room0student0': {
			'rulebook': ('physical', 'dorm2room0student0'),
			'location': 'dorm2room0'
		},
		'dorm0room4': {
			'rulebook': ('physical', 'dorm0room4')
		},
		'dorm0room4student1': {
			'rulebook': ('physical', 'dorm0room4student1'),
			'location': 'dorm0room4'
		},
		'dorm2room3': {
			'rulebook': ('physical', 'dorm2room3')
		},
		'dorm0room1': {
			'rulebook': ('physical', 'dorm0room1')
		},
		'dorm1room1student0': {
			'rulebook': ('physical', 'dorm1room1student0'),
			'location': 'dorm1room1'
		},
		'dorm0room0': {
			'rulebook': ('physical', 'dorm0room0')
		},
		'dorm1room3student0': {
			'rulebook': ('physical', 'dorm1room3student0'),
			'location': 'dorm1room3'
		},
		'dorm2room2': {
			'rulebook': ('physical', 'dorm2room2')
		},
		'dorm2room4student1': {
			'rulebook': ('physical', 'dorm2room4student1'),
			'location': 'dorm2room4'
		},
		'dorm0room5student0': {
			'rulebook': ('physical', 'dorm0room5student0'),
			'location': 'dorm0room5'
		},
		'dorm1room5': {
			'rulebook': ('physical', 'dorm1room5')
		},
		'dorm0room3': {
			'rulebook': ('physical', 'dorm0room3')
		},
		'dorm2room5student1': {
			'rulebook': ('physical', 'dorm2room5student1'),
			'location': 'dorm2room5'
		},
		'common1': {
			'rulebook': ('physical', 'common1')
		},
		'dorm2room4': {
			'rulebook': ('physical', 'dorm2room4')
		}
	},
	'hour': 0,
	'name': 'physical',
	'edge_val': {
		'common0': {
			'dorm0room5': {
				'is_mirror': True
			},
			'dorm0room3': {
				'is_mirror': True
			},
			'classroom': {},
			'dorm0room1': {
				'is_mirror': True
			},
			'dorm0room0': {
				'is_mirror': True
			},
			'dorm0room4': {
				'is_mirror': True
			},
			'dorm0room2': {
				'is_mirror': True
			}
		},
		'dorm1room3': {
			'common1': {}
		},
		'common2': {
			'dorm2room1': {
				'is_mirror': True
			},
			'dorm2room2': {
				'is_mirror': True
			},
			'dorm2room0': {
				'is_mirror': True
			},
			'classroom': {},
			'dorm2room5': {
				'is_mirror': True
			},
			'dorm2room4': {
				'is_mirror': True
			},
			'dorm2room3': {
				'is_mirror': True
			}
		},
		'classroom': {
			'common0': {
				'is_mirror': True
			},
			'common2': {
				'is_mirror': True
			},
			'common1': {
				'is_mirror': True
			}
		},
		'dorm1room1': {
			'common1': {}
		},
		'dorm2room0': {
			'common2': {}
		},
		'dorm1room0': {
			'common1': {}
		},
		'dorm0room0': {
			'common0': {}
		},
		'dorm0room4': {
			'common0': {}
		},
		'dorm2room3': {
			'common2': {}
		},
		'dorm0room1': {
			'common0': {}
		},
		'dorm2room2': {
			'common2': {}
		},
		'dorm1room5': {
			'common1': {}
		},
		'dorm1room2': {
			'common1': {}
		},
		'dorm0room2': {
			'common0': {}
		},
		'dorm2room1': {
			'common2': {}
		},
		'dorm2room5': {
			'common2': {}
		},
		'dorm2room4': {
			'common2': {}
		},
		'dorm0room3': {
			'common0': {}
		},
		'dorm0room5': {
			'common0': {}
		},
		'common1': {
			'dorm1room3': {
				'is_mirror': True
			},
			'dorm1room1': {
				'is_mirror': True
			},
			'classroom': {},
			'dorm1room5': {
				'is_mirror': True
			},
			'dorm1room0': {
				'is_mirror': True
			},
			'dorm1room2': {
				'is_mirror': True
			},
			'dorm1room4': {
				'is_mirror': True
			}
		},
		'dorm1room4': {
			'common1': {}
		}
	},
	'rulebooks': {
		'character': ('physical', 'character'),
		'thing': ('physical', 'character_thing'),
		'unit': ('physical', 'unit'),
		'place': ('physical', 'character_place'),
		'portal': ('physical', 'character_portal')
	}
}
DORM_INITIAL_COPY = {
	'rulebooks': {
		'thing': ('dorm0', 'character_thing'),
		'character': ('dorm0', 'character'),
		'unit': ('dorm0', 'unit'),
		'portal': ('dorm0', 'character_portal'),
		'place': ('dorm0', 'character_place')
	},
	'name': 'dorm0',
	'units': {
		'physical':
		frozenset({
			'dorm0room1', 'dorm0room2', 'dorm0room4', 'dorm0room5',
			'dorm0room3', 'common0', 'dorm0room0'
		})
	}
}
STUDENT_INITIAL_COPY = {
	'units': {
		'physical': frozenset({'dorm0room0student0'})
	},
	'xp': 0,
	'lazy': True,
	'drunkard': False,
	'name': 'dorm0room0student0',
	'rulebooks': {
		'place': ('dorm0room0student0', 'character_place'),
		'unit': ('dorm0room0student0', 'unit'),
		'portal': ('dorm0room0student0', 'character_portal'),
		'thing': ('dorm0room0student0', 'character_thing'),
		'character': ('dorm0room0student0', 'character')
	},
	'node_val': {
		'cell26': {
			'rulebook': ('dorm0room0student0', 'cell26'),
			'drunk': 0,
			'slow': 0
		},
		'cell13': {
			'rulebook': ('dorm0room0student0', 'cell13'),
			'drunk': 0,
			'slow': 0
		},
		'cell86': {
			'rulebook': ('dorm0room0student0', 'cell86'),
			'drunk': 0,
			'slow': 0
		},
		'cell11': {
			'rulebook': ('dorm0room0student0', 'cell11'),
			'drunk': 0,
			'slow': 0
		},
		'cell24': {
			'rulebook': ('dorm0room0student0', 'cell24'),
			'drunk': 0,
			'slow': 0
		},
		'cell53': {
			'rulebook': ('dorm0room0student0', 'cell53'),
			'drunk': 0,
			'slow': 0
		},
		'cell60': {
			'rulebook': ('dorm0room0student0', 'cell60'),
			'drunk': 0,
			'slow': 0
		},
		'cell55': {
			'rulebook': ('dorm0room0student0', 'cell55'),
			'drunk': 0,
			'slow': 0
		},
		'cell51': {
			'rulebook': ('dorm0room0student0', 'cell51'),
			'drunk': 0,
			'slow': 0
		},
		'cell25': {
			'rulebook': ('dorm0room0student0', 'cell25'),
			'drunk': 0,
			'slow': 0
		},
		'cell98': {
			'rulebook': ('dorm0room0student0', 'cell98'),
			'drunk': 0,
			'slow': 0
		},
		'cell85': {
			'rulebook': ('dorm0room0student0', 'cell85'),
			'drunk': 0,
			'slow': 0
		},
		'cell81': {
			'rulebook': ('dorm0room0student0', 'cell81'),
			'drunk': 0,
			'slow': 0
		},
		'cell63': {
			'rulebook': ('dorm0room0student0', 'cell63'),
			'drunk': 0,
			'slow': 0
		},
		'cell47': {
			'rulebook': ('dorm0room0student0', 'cell47'),
			'drunk': 0,
			'slow': 0
		},
		'cell78': {
			'rulebook': ('dorm0room0student0', 'cell78'),
			'drunk': 0,
			'slow': 0
		},
		'cell0': {
			'rulebook': ('dorm0room0student0', 'cell0'),
			'drunk': 0,
			'slow': 0
		},
		'cell12': {
			'rulebook': ('dorm0room0student0', 'cell12'),
			'drunk': 0,
			'slow': 0
		},
		'cell35': {
			'rulebook': ('dorm0room0student0', 'cell35'),
			'drunk': 0,
			'slow': 0
		},
		'cell99': {
			'rulebook': ('dorm0room0student0', 'cell99'),
			'drunk': 0,
			'slow': 0
		},
		'cell16': {
			'rulebook': ('dorm0room0student0', 'cell16'),
			'drunk': 0,
			'slow': 0
		},
		'cell57': {
			'rulebook': ('dorm0room0student0', 'cell57'),
			'drunk': 0,
			'slow': 0
		},
		'cell92': {
			'rulebook': ('dorm0room0student0', 'cell92'),
			'drunk': 0,
			'slow': 0
		},
		'cell1': {
			'rulebook': ('dorm0room0student0', 'cell1'),
			'drunk': 0,
			'slow': 0
		},
		'cell62': {
			'rulebook': ('dorm0room0student0', 'cell62'),
			'drunk': 0,
			'slow': 0
		},
		'cell80': {
			'rulebook': ('dorm0room0student0', 'cell80'),
			'drunk': 0,
			'slow': 0
		},
		'cell44': {
			'rulebook': ('dorm0room0student0', 'cell44'),
			'drunk': 0,
			'slow': 0
		},
		'cell82': {
			'rulebook': ('dorm0room0student0', 'cell82'),
			'drunk': 0,
			'slow': 0
		},
		'cell37': {
			'rulebook': ('dorm0room0student0', 'cell37'),
			'drunk': 0,
			'slow': 0
		},
		'cell20': {
			'rulebook': ('dorm0room0student0', 'cell20'),
			'drunk': 0,
			'slow': 0
		},
		'cell87': {
			'rulebook': ('dorm0room0student0', 'cell87'),
			'drunk': 0,
			'slow': 0
		},
		'cell64': {
			'rulebook': ('dorm0room0student0', 'cell64'),
			'drunk': 0,
			'slow': 0
		},
		'cell3': {
			'rulebook': ('dorm0room0student0', 'cell3'),
			'drunk': 0,
			'slow': 0
		},
		'cell23': {
			'rulebook': ('dorm0room0student0', 'cell23'),
			'drunk': 0,
			'slow': 0
		},
		'cell28': {
			'rulebook': ('dorm0room0student0', 'cell28'),
			'drunk': 0,
			'slow': 0
		},
		'cell95': {
			'rulebook': ('dorm0room0student0', 'cell95'),
			'drunk': 0,
			'slow': 0
		},
		'cell59': {
			'rulebook': ('dorm0room0student0', 'cell59'),
			'drunk': 0,
			'slow': 0
		},
		'cell91': {
			'rulebook': ('dorm0room0student0', 'cell91'),
			'drunk': 0,
			'slow': 0
		},
		'cell34': {
			'rulebook': ('dorm0room0student0', 'cell34'),
			'drunk': 0,
			'slow': 0
		},
		'cell32': {
			'rulebook': ('dorm0room0student0', 'cell32'),
			'drunk': 0,
			'slow': 0
		},
		'cell31': {
			'rulebook': ('dorm0room0student0', 'cell31'),
			'drunk': 0,
			'slow': 0
		},
		'cell21': {
			'rulebook': ('dorm0room0student0', 'cell21'),
			'drunk': 0,
			'slow': 0
		},
		'cell4': {
			'rulebook': ('dorm0room0student0', 'cell4'),
			'drunk': 0,
			'slow': 0
		},
		'cell77': {
			'rulebook': ('dorm0room0student0', 'cell77'),
			'drunk': 0,
			'slow': 0
		},
		'cell27': {
			'rulebook': ('dorm0room0student0', 'cell27'),
			'drunk': 0,
			'slow': 0
		},
		'cell61': {
			'rulebook': ('dorm0room0student0', 'cell61'),
			'drunk': 0,
			'slow': 0
		},
		'cell70': {
			'rulebook': ('dorm0room0student0', 'cell70'),
			'drunk': 0,
			'slow': 0
		},
		'cell84': {
			'rulebook': ('dorm0room0student0', 'cell84'),
			'drunk': 0,
			'slow': 0
		},
		'cell36': {
			'rulebook': ('dorm0room0student0', 'cell36'),
			'drunk': 0,
			'slow': 0
		},
		'cell30': {
			'rulebook': ('dorm0room0student0', 'cell30'),
			'drunk': 0,
			'slow': 0
		},
		'cell88': {
			'rulebook': ('dorm0room0student0', 'cell88'),
			'drunk': 0,
			'slow': 0
		},
		'cell52': {
			'rulebook': ('dorm0room0student0', 'cell52'),
			'drunk': 0,
			'slow': 0
		},
		'cell58': {
			'rulebook': ('dorm0room0student0', 'cell58'),
			'drunk': 0,
			'slow': 0
		},
		'cell2': {
			'rulebook': ('dorm0room0student0', 'cell2'),
			'drunk': 0,
			'slow': 0
		},
		'cell75': {
			'rulebook': ('dorm0room0student0', 'cell75'),
			'drunk': 0,
			'slow': 0
		},
		'cell69': {
			'rulebook': ('dorm0room0student0', 'cell69'),
			'drunk': 0,
			'slow': 0
		},
		'cell74': {
			'rulebook': ('dorm0room0student0', 'cell74'),
			'drunk': 0,
			'slow': 0
		},
		'cell49': {
			'rulebook': ('dorm0room0student0', 'cell49'),
			'drunk': 0,
			'slow': 0
		},
		'cell90': {
			'rulebook': ('dorm0room0student0', 'cell90'),
			'drunk': 0,
			'slow': 0
		},
		'cell15': {
			'rulebook': ('dorm0room0student0', 'cell15'),
			'drunk': 0,
			'slow': 0
		},
		'cell65': {
			'rulebook': ('dorm0room0student0', 'cell65'),
			'drunk': 0,
			'slow': 0
		},
		'cell50': {
			'rulebook': ('dorm0room0student0', 'cell50'),
			'drunk': 0,
			'slow': 0
		},
		'cell22': {
			'rulebook': ('dorm0room0student0', 'cell22'),
			'drunk': 0,
			'slow': 0
		},
		'cell46': {
			'rulebook': ('dorm0room0student0', 'cell46'),
			'drunk': 0,
			'slow': 0
		},
		'cell93': {
			'rulebook': ('dorm0room0student0', 'cell93'),
			'drunk': 0,
			'slow': 0
		},
		'cell40': {
			'rulebook': ('dorm0room0student0', 'cell40'),
			'drunk': 0,
			'slow': 0
		},
		'cell66': {
			'rulebook': ('dorm0room0student0', 'cell66'),
			'drunk': 0,
			'slow': 0
		},
		'cell73': {
			'rulebook': ('dorm0room0student0', 'cell73'),
			'drunk': 0,
			'slow': 0
		},
		'cell8': {
			'rulebook': ('dorm0room0student0', 'cell8'),
			'drunk': 0,
			'slow': 0
		},
		'cell41': {
			'rulebook': ('dorm0room0student0', 'cell41'),
			'drunk': 0,
			'slow': 0
		},
		'cell17': {
			'rulebook': ('dorm0room0student0', 'cell17'),
			'drunk': 0,
			'slow': 0
		},
		'cell56': {
			'rulebook': ('dorm0room0student0', 'cell56'),
			'drunk': 0,
			'slow': 0
		},
		'cell48': {
			'rulebook': ('dorm0room0student0', 'cell48'),
			'drunk': 0,
			'slow': 0
		},
		'cell89': {
			'rulebook': ('dorm0room0student0', 'cell89'),
			'drunk': 0,
			'slow': 0
		},
		'cell5': {
			'rulebook': ('dorm0room0student0', 'cell5'),
			'drunk': 0,
			'slow': 0
		},
		'cell94': {
			'rulebook': ('dorm0room0student0', 'cell94'),
			'drunk': 0,
			'slow': 0
		},
		'cell19': {
			'rulebook': ('dorm0room0student0', 'cell19'),
			'drunk': 0,
			'slow': 0
		},
		'cell71': {
			'rulebook': ('dorm0room0student0', 'cell71'),
			'drunk': 0,
			'slow': 0
		},
		'cell67': {
			'rulebook': ('dorm0room0student0', 'cell67'),
			'drunk': 0,
			'slow': 0
		},
		'cell96': {
			'rulebook': ('dorm0room0student0', 'cell96'),
			'drunk': 0,
			'slow': 0
		},
		'cell18': {
			'rulebook': ('dorm0room0student0', 'cell18'),
			'drunk': 0,
			'slow': 0
		},
		'cell42': {
			'rulebook': ('dorm0room0student0', 'cell42'),
			'drunk': 0,
			'slow': 0
		},
		'cell38': {
			'rulebook': ('dorm0room0student0', 'cell38'),
			'drunk': 0,
			'slow': 0
		},
		'cell29': {
			'rulebook': ('dorm0room0student0', 'cell29'),
			'drunk': 0,
			'slow': 0
		},
		'cell54': {
			'rulebook': ('dorm0room0student0', 'cell54'),
			'drunk': 0,
			'slow': 0
		},
		'cell14': {
			'rulebook': ('dorm0room0student0', 'cell14'),
			'drunk': 0,
			'slow': 0
		},
		'cell9': {
			'rulebook': ('dorm0room0student0', 'cell9'),
			'drunk': 0,
			'slow': 0
		},
		'cell33': {
			'rulebook': ('dorm0room0student0', 'cell33'),
			'drunk': 0,
			'slow': 0
		},
		'cell68': {
			'rulebook': ('dorm0room0student0', 'cell68'),
			'drunk': 0,
			'slow': 0
		},
		'cell10': {
			'rulebook': ('dorm0room0student0', 'cell10'),
			'drunk': 0,
			'slow': 0
		},
		'cell7': {
			'rulebook': ('dorm0room0student0', 'cell7'),
			'drunk': 0,
			'slow': 0
		},
		'cell76': {
			'rulebook': ('dorm0room0student0', 'cell76'),
			'drunk': 0,
			'slow': 0
		},
		'cell45': {
			'rulebook': ('dorm0room0student0', 'cell45'),
			'drunk': 0,
			'slow': 0
		},
		'cell83': {
			'rulebook': ('dorm0room0student0', 'cell83'),
			'drunk': 0,
			'slow': 0
		},
		'cell43': {
			'rulebook': ('dorm0room0student0', 'cell43'),
			'drunk': 0,
			'slow': 0
		},
		'cell79': {
			'rulebook': ('dorm0room0student0', 'cell79'),
			'drunk': 0,
			'slow': 0
		},
		'cell97': {
			'rulebook': ('dorm0room0student0', 'cell97'),
			'drunk': 0,
			'slow': 0
		},
		'cell39': {
			'rulebook': ('dorm0room0student0', 'cell39'),
			'drunk': 0,
			'slow': 0
		},
		'cell6': {
			'rulebook': ('dorm0room0student0', 'cell6'),
			'drunk': 0,
			'slow': 0
		},
		'cell72': {
			'rulebook': ('dorm0room0student0', 'cell72'),
			'drunk': 0,
			'slow': 0
		}
	}
}



from LiSE.examples.polygons import install
from LiSE import Engine


# TODO: use a test sim that does everything in every cache
def test_resume(tempdir):
	with Engine(tempdir) as eng:
		install(eng)
		eng.next_turn()
		last_branch, last_turn, last_tick = eng._btt()
	with Engine(tempdir) as eng:
		# tick gets bumped when we save the state of the randomizer
		assert eng._btt() == (last_branch, last_turn, last_tick + 1)
		curturn = eng.turn
		eng.next_turn()
		assert eng.turn == curturn + 1



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import cherrypy
import threading
import logging
from queue import Queue
from ..handle import EngineHandle


class LiSEHandleWebService(object):
	exposed = True

	def __init__(self, *args, **kwargs):
		if 'logger' in kwargs:
			self.logger = kwargs['logger']
		else:
			self.logger = kwargs['logger'] = logging.getLogger(__name__)
		self.cmdq = kwargs['cmdq'] = Queue()
		self.outq = kwargs['outq'] = Queue()
		self._handle_thread = threading.Thread(target=self._run_handle_forever,
												args=args,
												kwargs=kwargs,
												daemon=True)
		self._handle_thread.start()

	@staticmethod
	def _run_handle_forever(*args, **kwargs):
		cmdq = kwargs.pop('cmdq')
		outq = kwargs.pop('outq')
		logger = kwargs.pop('logger')
		setup = kwargs.pop('setup', None)
		logq = Queue()

		def log(typ, data):
			if typ == 'command':
				(cmd, args) = data
				logger.debug("LiSE thread {}: calling {}{}".format(
					threading.get_ident(), cmd, tuple(args)))
			else:
				logger.debug(
					"LiSE thread {}: returning {} (of type {})".format(
						threading.get_ident(), data, repr(type(data))))

		def get_log_forever(logq):
			(level, data) = logq.get()
			getattr(logger, level)(data)

		engine_handle = EngineHandle(args, kwargs, logq)
		if setup:
			setup(engine_handle._real)
		handle_log_thread = threading.Thread(target=get_log_forever,
												args=(logq, ),
												daemon=True)
		handle_log_thread.start()
		while True:
			inst = cmdq.get()
			if inst == 'shutdown':
				handle_log_thread.join()
				cmdq.close()
				outq.close()
				return 0
			cmd = inst.pop('command')
			silent = inst.pop('silent', False)
			log('command', (cmd, args))
			response = getattr(engine_handle, cmd)(**inst)
			if silent:
				continue
			log('result', response)
			outq.put(engine_handle._real.listify(response))

	@cherrypy.tools.accept(media='application/json')
	@cherrypy.tools.json_out()
	def GET(self):
		return cherrypy.session['LiSE_response']

	@cherrypy.tools.json_out()
	def POST(self, **kwargs):
		silent = kwargs.get('silent', False)
		self.cmdq.put(kwargs)
		if silent:
			return None
		response = self.outq.get()
		cherrypy.session['LiSE_response'] = response
		return response

	def PUT(self, silent=False, **kwargs):
		silent = silent
		self.cmdq.put(kwargs)
		if not silent:
			cherrypy.session['LiSE_response'] = self.outq.get()

	def DELETE(self):
		cherrypy.session.pop('LiSE_response', None)



# This file is part of allegedb, an object-relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""WindowDict, the core data structure used by allegedb's caching system.

It resembles a dictionary, more specifically a defaultdict-like where retrieving
a key that isn't set will get the highest set key that is lower than the key
you asked for (and thus, keys must be orderable). It is optimized for retrieval
of the same key and neighboring ones repeatedly and in sequence.

"""
from abc import abstractmethod, ABC
from collections import deque
from collections.abc import Mapping, MutableMapping, KeysView, ItemsView, ValuesView
from itertools import chain
from operator import itemgetter, lt, le
from typing import Union, Callable, Dict, List, Tuple, Any, Iterable, Set, Optional
from enum import Enum

get0 = itemgetter(0)
get1 = itemgetter(1)


class Direction(Enum):
	FORWARD = 'forward'
	BACKWARD = 'backward'


# TODO: cancel changes that would put something back to where it was at the start
# This will complicate the update_window functions though, and I don't think it'll
# improve much apart from a bit of efficiency in that the deltas are smaller
# sometimes.
def update_window(turn_from: int, tick_from: int, turn_to: int, tick_to: int,
					updfun: Callable, branchd: Dict[int, List[tuple]]):
	"""Iterate over a window of time in ``branchd`` and call ``updfun`` on the values"""
	if turn_from in branchd:
		# Not including the exact tick you started from because deltas are *changes*
		for past_state in branchd[turn_from][tick_from + 1:]:
			updfun(*past_state)
	for midturn in range(turn_from + 1, turn_to):
		if midturn in branchd:
			for past_state in branchd[midturn][:]:
				updfun(*past_state)
	if turn_to in branchd:
		for past_state in branchd[turn_to][:tick_to + 1]:
			updfun(*past_state)


def update_backward_window(turn_from: int, tick_from: int, turn_to: int,
							tick_to: int, updfun: Callable,
							branchd: Dict[int, List[tuple]]):
	"""Iterate backward over a window of time in ``branchd`` and call ``updfun`` on the values"""
	if turn_from in branchd:
		for future_state in reversed(branchd[turn_from][:tick_from]):
			updfun(*future_state)
	for midturn in range(turn_from - 1, turn_to, -1):
		if midturn in branchd:
			for future_state in reversed(branchd[midturn][:]):
				updfun(*future_state)
	if turn_to in branchd:
		for future_state in reversed(branchd[turn_to][tick_to + 1:]):
			updfun(*future_state)


class HistoricKeyError(KeyError):
	"""KeyError subclass that distinguishes deleted keys from those that were never set"""

	def __init__(self, *args, deleted=False):
		super().__init__(*args)
		self.deleted = deleted


def within_history(rev: int, windowdict: 'WindowDict'):
	"""Return whether the windowdict has history at the revision."""
	if not windowdict:
		return False
	begin = windowdict._past[0][0] if windowdict._past else windowdict._future[
		-1][0]
	end = windowdict._future[0][0] if windowdict._future else windowdict._past[
		-1][0]
	return begin <= rev <= end


class WindowDictKeysView(ABC, KeysView):
	"""Look through all the keys a WindowDict contains."""
	_mapping: 'WindowDict'

	def __contains__(self, rev: int):
		return rev in self._mapping._keys

	def __iter__(self):
		past = self._mapping._past
		future = self._mapping._future
		if past:
			yield from map(get0, past)
		if future:
			yield from map(get0, reversed(future))


class WindowDictItemsView(ABC, ItemsView):
	"""Look through everything a WindowDict contains."""
	_mapping: 'WindowDict'

	def __contains__(self, item: Tuple[int, Any]):
		(rev, v) = item
		mapp = self._mapping
		if not within_history(rev, mapp):
			return False
		for mrev, mv in mapp._past:
			if mrev == rev:
				return mv == v
		for mrev, mv in mapp._future:
			if mrev == rev:
				return mv == v
		return False

	def __iter__(self):
		past = self._mapping._past
		future = self._mapping._future
		if past:
			yield from past
		if future:
			yield from future


class WindowDictPastFutureKeysView(ABC, KeysView):
	"""View on a WindowDict's keys relative to last lookup"""
	_mapping: Union['WindowDictPastView', 'WindowDictFutureView']

	def __iter__(self):
		if not self._mapping.stack:
			return
		yield from map(get0, reversed(self._mapping.stack))

	def __contains__(self, item: int):
		return item in self._mapping and item in map(get0, self._mapping.stack)


class WindowDictPastFutureItemsView(ABC, ItemsView):
	_mapping: Union['WindowDictPastView', 'WindowDictFutureView']

	@staticmethod
	@abstractmethod
	def _out_of_range(item: tuple, stack: list):
		pass

	def __iter__(self):
		if not self._mapping.stack:
			return
		yield from reversed(self._mapping.stack)

	def __contains__(self, item: Tuple[int, Any]):
		if self._out_of_range(item, self._mapping.stack):
			return False
		i0, i1 = item
		for j0, j1 in self._mapping.stack:
			if i0 == j0:
				return i1 == j1
		return False


class WindowDictPastItemsView(WindowDictPastFutureItemsView):

	@staticmethod
	def _out_of_range(item: Tuple[int, Any], stack: List[Tuple[int, Any]]):
		return item[0] < stack[0][0] or item[0] > stack[-1][0]


class WindowDictFutureItemsView(WindowDictPastFutureItemsView):
	"""View on a WindowDict's future items relative to last lookup"""

	@staticmethod
	def _out_of_range(item: Tuple[int, Any], stack: List[Tuple[int, Any]]):
		return item[0] < stack[-1][0] or item[0] > stack[0][0]


class WindowDictPastFutureValuesView(ABC, ValuesView):
	"""Abstract class for views on the past or future values of a WindowDict"""
	_mapping: Union['WindowDictPastView', 'WindowDictFutureView']

	def __iter__(self):
		stack = self._mapping.stack
		if not stack:
			return
		yield from map(get1, reversed(stack))

	def __contains__(self, item: Any):
		stack = self._mapping.stack
		if not stack:
			return False
		return item in map(get1, stack)


class WindowDictValuesView(ABC, ValuesView):
	"""Look through all the values that a WindowDict contains."""
	_mapping: 'WindowDict'

	def __contains__(self, value: Any):
		past = self._mapping._past
		future = self._mapping._future
		if past:
			for rev, v in past:
				if v == value:
					return True
		if future:
			for rev, v in future:
				if v == value:
					return True
		return False

	def __iter__(self):
		past = self._mapping._past
		future = self._mapping._future
		if past:
			yield from map(get1, past)
		if future:
			yield from map(get1, future)


class WindowDictPastFutureView(ABC, Mapping):
	"""Abstract class for historical views on WindowDict"""
	__slots__ = ('stack', )
	stack: List[Tuple[int, Any]]

	def __init__(self, stack: List[Tuple[int, Any]]) -> None:
		self.stack = stack

	def __len__(self) -> int:
		stack = self.stack
		if not stack:
			return 0
		return len(stack)


class WindowDictPastView(WindowDictPastFutureView):
	"""Read-only mapping of just the past of a WindowDict"""

	def __iter__(self) -> Iterable[int]:
		stack = self.stack
		if not stack:
			return
		yield from map(get0, reversed(stack))

	def __getitem__(self, key: int) -> Any:
		stack = self.stack
		if not stack or key < stack[0][0] or key > stack[-1][0]:
			raise KeyError
		for rev, value in stack:
			if rev == key:
				return value
		raise KeyError

	def keys(self) -> WindowDictPastFutureKeysView:
		return WindowDictPastFutureKeysView(self)

	def items(self) -> WindowDictPastItemsView:
		return WindowDictPastItemsView(self)

	def values(self) -> WindowDictPastFutureValuesView:
		return WindowDictPastFutureValuesView(self)


class WindowDictFutureView(WindowDictPastFutureView):
	"""Read-only mapping of just the future of a WindowDict"""

	def __iter__(self):
		stack = self.stack
		if not stack:
			return
		yield from map(get0, reversed(stack))

	def __getitem__(self, key: int):
		stack = self.stack
		if not stack or key < stack[-1][0] or key > stack[0][0]:
			raise KeyError
		for rev, value in stack:
			if rev == key:
				return value
		raise KeyError

	def keys(self) -> WindowDictPastFutureKeysView:
		return WindowDictPastFutureKeysView(self)

	def items(self) -> WindowDictFutureItemsView:
		return WindowDictFutureItemsView(self)

	def values(self) -> WindowDictPastFutureValuesView:
		return WindowDictPastFutureValuesView(self)


class WindowDictSlice:
	"""A slice of history in which the start is earlier than the stop"""
	__slots__ = ['dic', 'slic']
	dic: 'WindowDict'
	slic: slice

	def __init__(self, dic: 'WindowDict', slic: slice):
		self.dic = dic
		self.slic = slic

	def __reversed__(self) -> Iterable[Any]:
		return iter(WindowDictReverseSlice(self.dic, self.slic))

	def __iter__(self):
		dic = self.dic
		if not dic:
			return
		slic = self.slic
		if slic.step is not None:
			for i in range(slic.start or dic.beginning, slic.stop
							or dic.end + 1, slic.step):
				yield dic[i]
			return
		if slic.start is None and slic.stop is None:
			yield from map(get1, dic._past)
			yield from map(get1, reversed(dic._future))
		elif slic.start is not None and slic.stop is not None:
			if slic.stop == slic.start:
				yield dic[slic.stop]
				return
			past = dic._past
			future = dic._future
			if slic.start < slic.stop:
				left, right = slic.start, slic.stop
				dic.seek(right)
				if not past:
					return
				if past[-1][0] == right:
					future.append(past.pop())
				cmp = lt
			else:
				left, right = slic.stop, slic.start
				dic.seek(right)
				if not past:
					return
				cmp = le
			it = iter(past)
			p0, p1 = next(it)
			while cmp(p0, left):
				try:
					p0, p1 = next(it)
				except StopIteration:
					return
			else:
				yield p1
			yield from map(get1, it)
		elif slic.start is None:
			stac = dic._past + list(reversed(dic._future))
			while stac and stac[-1][0] >= slic.stop:
				stac.pop()
			yield from map(get1, stac)
			return
		else:  # slic.stop is None
			if not dic._past and not dic._future:
				return
			chan = chain(dic._past, reversed(dic._future))
			nxt = next(chan)
			while nxt[0] < slic.start:
				try:
					nxt = next(chan)
				except StopIteration:
					return
			yield get1(nxt)
			yield from map(get1, chan)


class WindowDictReverseSlice:
	"""A slice of history in which the start is later than the stop"""
	__slots__ = ['dict', 'slice']

	def __init__(self, dict: 'WindowDict', slic: slice):
		self.dict = dict
		self.slice = slic

	def __reversed__(self):
		return iter(WindowDictSlice(self.dict, self.slice))

	def __iter__(self):
		dic = self.dict
		if not dic:
			return
		slic = self.slice
		if slic.step is not None:
			for i in range(slic.start or dic.end, slic.stop or dic.beginning,
							slic.step):
				yield dic[i]
			return
		if slic.start is None and slic.stop is None:
			yield from map(get1, dic._future)
			yield from map(get1, reversed(dic._past))
		elif slic.start is not None and slic.stop is not None:
			if slic.start == slic.stop:
				yield dic[slic.stop]
				return
			if slic.start < slic.stop:
				left, right = slic.start, slic.stop
				dic.seek(right)
				it = reversed(dic._past)
				next(it)
				cmp = lt
			else:
				left, right = slic.stop, slic.start
				dic.seek(right)
				it = reversed(dic._past)
				cmp = le
			for frev, fv in it:
				if cmp(frev, left):
					return
				yield fv
		elif slic.start is None:
			stac = dic._past + list(reversed(dic._future))
			while stac and stac[-1][0] >= slic.stop:
				stac.pop()
			yield from map(get1, reversed(stac))
		else:  # slic.stop is None
			stac = deque(dic._past)
			stac.extend(reversed(dic._future))
			while stac and stac[0][0] < slic.start:
				stac.popleft()
			yield from map(get1, reversed(stac))


class WindowDict(MutableMapping):
	"""A dict that keeps every value that a variable has had over time.

    Look up a revision number in this dict and it will give you the
    effective value as of that revision. Keys should always be
    revision numbers.

    Optimized for the cases where you look up the same revision
    repeatedly, or its neighbors.

    This supports slice notation to get all values in a given
    time-frame. If you do not supply a step, you'll just get the
    values, with no indication of when they're from exactly --
    so explicitly supply a step of 1 to get the value at each point in
    the slice, or use the ``future`` and ``past`` methods to get read-only
    mappings of data relative to a particular revision.

    Unlike slices of eg. lists, you can slice with a start greater than the stop
    even if you don't supply a step. That will get you values in reverse order.

    """
	__slots__ = ('_future', '_past', '_keys', 'beginning', 'end', '_last')

	_past: List[Tuple[int, Any]]
	_future: List[Tuple[int, Any]]
	_keys: Set[int]
	_last: Optional[int]
	beginning: Optional[int]
	end: Optional[int]

	def future(self, rev: int = None) -> WindowDictFutureView:
		"""Return a Mapping of items after the given revision.

        Default revision is the last one looked up.

        """
		if rev is not None:
			self.seek(rev)
		return WindowDictFutureView(self._future)

	def past(self, rev: int = None) -> WindowDictPastView:
		"""Return a Mapping of items at or before the given revision.

        Default revision is the last one looked up.

        """
		if rev is not None:
			self.seek(rev)
		return WindowDictPastView(self._past)

	def seek(self, rev: int) -> None:
		"""Arrange the caches to help look up the given revision."""
		# TODO: binary search? Perhaps only when one or the other
		#       stack is very large?
		if rev == self._last:
			return
		past = self._past
		future = self._future
		if future:
			appender = past.append
			popper = future.pop
			future_start = future[-1][0]
			while future_start <= rev:
				appender(popper())
				if future:
					future_start = future[-1][0]
				else:
					break
		if past:
			popper = past.pop
			appender = future.append
			past_end = past[-1][0]
			while past_end > rev:
				appender(popper())
				if past:
					past_end = past[-1][0]
				else:
					break
		self._last = rev

	def rev_gettable(self, rev: int) -> bool:
		beg = self.beginning
		if beg is None:
			return False
		return rev >= beg

	def rev_before(self, rev: int):
		"""Return the latest past rev on which the value changed."""
		self.seek(rev)
		if self._past:
			return self._past[-1][0]

	def rev_after(self, rev: int):
		"""Return the earliest future rev on which the value will change."""
		self.seek(rev)
		if self._future:
			return self._future[-1][0]

	def initial(self) -> Any:
		"""Return the earliest value we have"""
		if self._past:
			return self._past[0][1]
		if self._future:
			return self._future[-1][1]
		raise KeyError("No data")

	def final(self) -> Any:
		"""Return the latest value we have"""
		if self._future:
			return self._future[0][1]
		if self._past:
			return self._past[-1][1]
		raise KeyError("No data")

	def truncate(self, rev: int, direction: Direction = 'forward') -> None:
		"""Delete everything after the given revision, exclusive.

        With direction='backward', delete everything before the revision,
        exclusive, instead.

        """
		self.seek(rev)
		if direction == 'forward':
			self._keys.difference_update(map(get0, self._future))
			self._future = []
			if not self._past:
				self.beginning = self.end = None
		elif direction == 'backward':
			if not self._past:
				return
			self._keys.difference_update(map(get0, self._past[:-1]))
			self._past = [self._past[-1]]
			if not self._future:
				self.beginning = self.end = None
		else:
			raise ValueError("Need direction 'forward' or 'backward'")

	def keys(self) -> WindowDictKeysView:
		return WindowDictKeysView(self)

	def items(self) -> WindowDictItemsView:
		return WindowDictItemsView(self)

	def values(self) -> WindowDictValuesView:
		return WindowDictValuesView(self)

	def __bool__(self) -> bool:
		return bool(self._past) or bool(self._future)

	def __init__(
			self,
			data: Union[List[Tuple[int, Any]], Dict[int, Any]] = None) -> None:
		if not data:
			self._past = []
		elif isinstance(data, Mapping):
			self._past = list(data.items())
		else:
			# assume it's an orderable sequence of pairs
			self._past = list(data)
		self._past.sort()
		self._future = []
		self._keys = set(map(get0, self._past))
		self.beginning = None if not self._past else self._past[0][0]
		self.end = None if not self._past else self._past[-1][0]
		self._last = None

	def __iter__(self) -> Iterable[Any]:
		if not self:
			return
		if self._past:
			yield from map(get0, self._past)
		if self._future:
			yield from map(get0, self._future)

	def __contains__(self, item: int) -> bool:
		return item in self._keys

	def __len__(self) -> int:
		return len(self._past) + len(self._future)

	def __getitem__(self, rev: int) -> Any:
		if isinstance(rev, slice):
			if None not in (rev.start, rev.stop) and rev.start > rev.stop:
				return WindowDictReverseSlice(self, rev)
			return WindowDictSlice(self, rev)
		self.seek(rev)
		past = self._past
		if not past:
			raise HistoricKeyError(
				"Revision {} is before the start of history".format(rev))
		return past[-1][1]

	def __setitem__(self, rev: int, v: Any) -> None:
		past = self._past
		if past or self._future:
			self.seek(rev)
			if past:
				if past[-1][0] == rev:
					past[-1] = (rev, v)
				else:
					past.append((rev, v))
			else:
				past.append((rev, v))
				self.beginning = rev
			end = self.end
			if end is None or rev > end:
				self.end = rev
		else:
			past.append((rev, v))
			self.beginning = self.end = self._last = rev
		self._keys.add(rev)

	def __delitem__(self, rev: int) -> None:
		# Not checking for rev's presence at the beginning because
		# to do so would likely require iterating thru history,
		# which I have to do anyway in deleting.
		# But handle degenerate case.
		if not self:
			raise HistoricKeyError("Tried to delete from an empty WindowDict")
		if self.beginning is None:
			if self.end is not None and rev > self.end:
				raise HistoricKeyError(
					"Rev outside of history: {}".format(rev))
		elif self.end is None:
			if self.beginning is not None and rev < self.beginning:
				raise HistoricKeyError(
					"Rev outside of history: {}".format(rev))
		elif not self.beginning <= rev <= self.end:
			raise HistoricKeyError("Rev outside of history: {}".format(rev))
		self.seek(rev)
		past = self._past
		future = self._future
		if not past or past[-1][0] != rev:
			raise HistoricKeyError("Rev not present: {}".format(rev))
		del past[-1]
		if not past:
			if future:
				self.beginning = future[-1][0]
			else:
				self.beginning = self.end = None
		elif not future:
			self.end = past[-1][0]
		self._keys.remove(rev)

	def __repr__(self) -> str:
		me = dict(self._past)
		me.update(self._future)
		return "{}({})".format(self.__class__.__name__, me)


class FuturistWindowDict(WindowDict):
	"""A WindowDict that does not let you rewrite the past."""
	__slots__ = ('_future', '_past', 'beginning')
	_future: List[Tuple[int, Any]]
	_past: List[Tuple[int, Any]]
	beginning: Optional[int]

	def __setitem__(self, rev: int, v: Any) -> None:
		if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap'):
			v = v.unwrap()
		self.seek(rev)
		past = self._past
		future = self._future
		if future:
			raise HistoricKeyError(
				"Already have some history after {}".format(rev))
		if not past:
			self.beginning = rev
			past.append((rev, v))
		elif rev > past[-1][0]:
			past.append((rev, v))
		elif rev == past[-1][0]:
			past[-1] = (rev, v)
		else:
			raise HistoricKeyError(
				"Already have some history after {} "
				"(and my seek function is broken?)".format(rev))
		if self.end is None or rev > self.end:
			self.end = rev
		self._keys.add(rev)


class TurnDict(FuturistWindowDict):
	__slots__ = ('_future', '_past')
	_future: List[Tuple[int, Any]]
	_past: List[Tuple[int, Any]]
	cls = FuturistWindowDict

	def __setitem__(self, turn: int, value: Any) -> None:
		if type(value) is not FuturistWindowDict:
			value = FuturistWindowDict(value)
		FuturistWindowDict.__setitem__(self, turn, value)


class SettingsTurnDict(WindowDict):
	__slots__ = ('_future', '_past')
	_future: List[Tuple[int, Any]]
	_past: List[Tuple[int, Any]]
	cls = WindowDict

	def __setitem__(self, turn: int, value: Any) -> None:
		if type(value) is not WindowDict:
			value = WindowDict(value)
		WindowDict.__setitem__(self, turn, value)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .allegedb.cache import (Cache, PickyDefaultDict, StructuredDefaultDict,
								TurnDict, WindowDict, HistoricKeyError)
from .util import singleton_get, sort_set
from collections import OrderedDict


class InitializedCache(Cache):
	__slots__ = ()

	def _store_journal(self, *args):
		entity, key, branch, turn, tick, value = args[-6:]
		parent = args[:-6]
		settings_turns = self.settings[branch]
		presettings_turns = self.presettings[branch]
		try:
			prev = self.retrieve(*args[:-1])
		except KeyError:
			return  # because you can't rewind past this
		if prev == value:
			return  # not much point reporting on a non-change in a diff
		if turn in settings_turns or turn in settings_turns.future():
			assert turn in presettings_turns or turn in presettings_turns.future(
			)
			setticks = settings_turns[turn]
			presetticks = presettings_turns[turn]
			presetticks[tick] = parent + (entity, key, prev)
			setticks[tick] = parent + (entity, key, value)
		else:
			presettings_turns[turn] = {tick: parent + (entity, key, prev)}
			settings_turns[turn] = {tick: parent + (entity, key, value)}


class EntitylessCache(Cache):
	__slots__ = ()

	def store(self,
				key,
				branch,
				turn,
				tick,
				value,
				*,
				planning=None,
				forward=None,
				loading=False,
				contra=True):
		super().store(None,
						key,
						branch,
						turn,
						tick,
						value,
						planning=planning,
						forward=forward,
						loading=loading,
						contra=contra)

	def iter_entities_or_keys(self, branch, turn, tick, *, forward=None):
		return super().iter_entities_or_keys(None,
												branch,
												turn,
												tick,
												forward=forward)

	iter_entities = iter_keys = iter_entities_or_keys

	def contains_entity_or_key(self, ke, branch, turn, tick):
		return super().contains_entity_or_key(None, ke, branch, turn, tick)

	contains_entity = contains_key = contains_entity_or_key

	def retrieve(self, *args):
		return super().retrieve(*(None, ) + args)


class InitializedEntitylessCache(EntitylessCache, InitializedCache):
	__slots__ = ()


class UnitnessCache(Cache):
	"""A cache for remembering when a node is a unit of a character."""
	__slots__ = ('user_order', 'user_shallow', 'graphs', 'graph_units',
					'char_units', 'solo_unit', 'unique_unit', 'unique_graph',
					'users')

	def __init__(self, engine):
		Cache.__init__(self, engine)
		self.user_order = StructuredDefaultDict(3, TurnDict)
		self.user_shallow = PickyDefaultDict(TurnDict)
		self.graphs = StructuredDefaultDict(1, TurnDict)
		self.graph_units = StructuredDefaultDict(1, TurnDict)
		self.char_units = StructuredDefaultDict(1, TurnDict)
		self.solo_unit = StructuredDefaultDict(1, TurnDict)
		self.unique_unit = StructuredDefaultDict(1, TurnDict)
		self.unique_graph = StructuredDefaultDict(1, TurnDict)
		self.users = StructuredDefaultDict(1, TurnDict)

	def store(self,
				character,
				graph,
				node,
				branch,
				turn,
				tick,
				is_unit,
				*,
				planning=None,
				forward=None,
				loading=False,
				contra=True):
		is_unit = True if is_unit else None
		super().store(character,
						graph,
						node,
						branch,
						turn,
						tick,
						is_unit,
						planning=planning,
						forward=forward,
						loading=loading,
						contra=contra)
		userturns = self.user_order[graph][node][character][branch]
		if turn in userturns:
			userturns[turn][tick] = is_unit
		else:
			userturns[turn] = {tick: is_unit}
		usershal = self.user_shallow[(graph, node, character, branch)]
		if turn in usershal:
			usershal[turn][tick] = is_unit
		else:
			usershal[turn] = {tick: is_unit}
		charavs = self.char_units[character][branch]
		graphavs = self.graph_units[(character, graph)][branch]
		graphs = self.graphs[character][branch]
		uniqgraph = self.unique_graph[character][branch]
		soloav = self.solo_unit[(character, graph)][branch]
		uniqav = self.unique_unit[character][branch]
		users = self.users[graph, node][branch]

		def add_something(cache, what):
			if turn in cache:
				nucache = cache[turn][tick].copy()
				nucache.add(what)
				cache[turn][tick] = nucache
			elif cache.rev_gettable(turn):
				cacheturn = cache[turn]
				nucache = cacheturn[cacheturn.end].copy()
				nucache.add(what)
				cache[turn] = {tick: nucache}
			else:
				cache[turn] = {tick: {what}}

		def remove_something(cache, what):
			if turn in cache:
				nucache = cache[turn][tick].copy()
				nucache.remove(what)
				cache[turn][tick] = nucache
			elif cache.rev_gettable(turn):
				cacheturn = cache[turn]
				nucache = cacheturn[cacheturn.end].copy()
				nucache.remove(what)
				cache[turn] = {tick: nucache}
			else:
				raise ValueError

		if is_unit:
			add_something(graphavs, node)
			add_something(charavs, (graph, node))
			add_something(graphs, graph)
			add_something(users, character)
		else:
			remove_something(graphavs, node)
			remove_something(charavs, (graph, node))
			if not graphavs[turn][tick]:
				remove_something(graphs, graph)
			if not charavs[turn][tick]:
				remove_something(users, character)
		graphav = singleton_get(graphavs[turn][tick])
		if turn in soloav:
			soloav[turn][tick] = graphav
		else:
			soloav[turn] = {tick: graphav}
		charav = singleton_get(charavs[turn][tick])
		if turn in uniqav:
			uniqav[turn][tick] = charav
		else:
			uniqav[turn] = {tick: charav}
		if not graphavs[turn][tick]:
			graphs[turn][tick].remove(graph)
			if len(graphs[turn][tick]) == 1:
				uniqgraph[turn][tick] = next(iter(graphs[turn][tick]))
			else:
				uniqgraph[turn][tick] = None
		if turn in graphavs and tick in graphavs[turn] and len(
			graphavs[turn][tick]) != 1:
			if turn in soloav:
				soloav[turn][tick] = None
			else:
				soloav[turn] = {tick: None}
		else:
			if turn in soloav:
				soloav[turn][tick] = node
			else:
				soloav[turn] = {tick: None}
		if turn in charavs and charavs[turn].rev_gettable(tick) and len(
			charavs[turn][tick]) != 1:
			if turn in uniqav:
				uniqav[turn][tick] = None
			else:
				uniqav[turn] = {tick: None}
		elif turn in uniqav:
			uniqav[turn][tick] = (graph, node)
		else:
			uniqav[turn] = {tick: (graph, node)}
		if turn in graphs and graphs[turn].rev_gettable(tick) and len(
			graphs[turn][tick]) != 1:
			if turn in uniqgraph:
				uniqgraph[turn][tick] = None
			else:
				uniqgraph[turn] = {tick: None}
		elif turn in uniqgraph:
			uniqgraph[turn][tick] = graph
		else:
			uniqgraph[turn] = {tick: graph}

	def get_char_graph_avs(self, char, graph, branch, turn, tick):
		return self._valcache_lookup(self.graph_units[(char, graph)], branch,
										turn, tick) or set()

	def get_char_graph_solo_av(self, char, graph, branch, turn, tick):
		return self._valcache_lookup(self.solo_unit[(char, graph)], branch,
										turn, tick)

	def get_char_only_av(self, char, branch, turn, tick):
		return self._valcache_lookup(self.unique_unit[char], branch, turn,
										tick)

	def get_char_only_graph(self, char, branch, turn, tick):
		return self._valcache_lookup(self.unique_graph[char], branch, turn,
										tick)

	def get_char_graphs(self, char, branch, turn, tick):
		return self._valcache_lookup(self.graphs[char], branch, turn,
										tick) or set()

	def _slow_iter_character_avatars(self, character, branch, turn, tick, *,
										forward):
		for graph in self.iter_entities(character,
										branch,
										turn,
										tick,
										forward=forward):
			for node in self.iter_entities(character,
											graph,
											branch,
											turn,
											tick,
											forward=forward):
				yield graph, node

	def _slow_iter_users(self, graph, node, branch, turn, tick):
		if graph not in self.user_order:
			return
		for character in self.user_order[graph][node]:
			if (graph, node, character, branch) not in self.user_shallow:
				for (b, t, tc) in self.db._iter_parent_btt(branch, turn, tick):
					if b in self.user_order[graph][node][character]:
						isav = self.user_order[graph][node][character][b][t]
						# side effect!! bad!
						self.store(character, graph, node, branch, turn, tick,
									isav[tc])
						break
				else:
					self.store(character, graph, node, branch, turn, tick,
								None)
			try:
				if self.user_shallow[(graph, node, character,
										branch)][turn][tick]:
					yield character
			except HistoricKeyError:
				continue


class RulesHandledCache(object):

	def __init__(self, engine):
		self.engine = engine
		self.handled = {}
		self.handled_deep = StructuredDefaultDict(1, type=WindowDict)
		self.unhandled = {}

	def get_rulebook(self, *args):
		raise NotImplementedError

	def iter_unhandled_rules(self, branch, turn, tick):
		raise NotImplementedError

	def store(self, *args, loading=False):
		entity = args[:-5]
		rulebook, rule, branch, turn, tick = args[-5:]
		self.handled.setdefault(entity + (rulebook, branch, turn),
								set()).add(rule)
		self.handled_deep[branch][turn][tick] = (entity, rulebook, rule)
		unhandl = self.unhandled.setdefault(entity, {}).setdefault(
			rulebook, {}).setdefault(branch, {})
		if turn not in unhandl:
			unhandl[turn] = list(
				self.unhandled_rulebook_rules(entity, rulebook, branch, turn,
												tick))
		try:
			unhandl[turn].remove(rule)
		except ValueError:
			if not loading:
				raise

	def fork(self, branch, turn, tick):
		parent_branch, parent_turn, parent_tick, end_turn, end_tick = self.engine._branches[
			branch]
		unhandl = self.unhandled
		handl = self.handled
		rbcache = self.engine._rulebooks_cache
		unhandrr = self.unhandled_rulebook_rules
		getrb = self.get_rulebook
		for entity in unhandl:
			rulebook = getrb(*entity + (branch, turn, tick))
			if entity + (rulebook, branch, turn) in handl:
				raise HistoricKeyError(
					"Tried to fork history in a RulesHandledCache, "
					"but it seems like rules have already been run where we're "
					"forking to")
			rules = set(rbcache.retrieve(rulebook, branch, turn, tick))
			unhandled_rules = unhandrr(entity, rulebook, parent_branch,
										parent_turn, tick)
			unhandled_rules_set = set(unhandled_rules)
			handl[entity + (rulebook, branch,
							turn)] = rules.difference(unhandled_rules_set)
			unhandl[entity][rulebook][branch][turn] = unhandled_rules

	def retrieve(self, *args):
		return self.handled[args]

	def unhandled_rulebook_rules(self, *args):
		entity = args[:-4]
		rulebook, branch, turn, tick = args[-4:]
		unh = self.unhandled
		if entity in unh:
			ue = unh[entity]
			if rulebook in ue:
				uer = ue[rulebook]
				if branch in uer:
					uerb = uer[branch]
					if turn in uerb:
						return uerb[turn]
		try:
			rulebook_rules = self.engine._rulebooks_cache.retrieve(
				rulebook, branch, turn, tick)
		except KeyError:
			return []
		handled_rules = self.handled.setdefault(
			entity + (rulebook, branch, turn), set())
		return [rule for rule in rulebook_rules if rule not in handled_rules]


class CharacterRulesHandledCache(RulesHandledCache):

	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_rulebooks_cache.retrieve(
				character, branch, turn, tick)
		except KeyError:
			return character, 'character'

	def iter_unhandled_rules(self, branch, turn, tick):
		for character in sort_set(self.engine.character.keys()):
			rb = self.get_rulebook(character, branch, turn, tick)
			for rule in self.unhandled_rulebook_rules(character, rb, branch,
														turn, tick):
				yield character, rb, rule


class UnitRulesHandledCache(RulesHandledCache):

	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._units_rulebooks_cache.retrieve(
				character, branch, turn, tick)
		except KeyError:
			return character, 'unit'

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			rulebook = self.get_rulebook(character, branch, turn, tick)
			charavm = charm[character].unit
			for graph in sort_set(charavm.keys()):
				for avatar in sort_set(charavm[graph].keys()):
					try:
						rules = self.unhandled_rulebook_rules(
							character, graph, avatar, rulebook, branch, turn,
							tick)
					except KeyError:
						continue
					for rule in rules:
						yield character, graph, avatar, rulebook, rule


class CharacterThingRulesHandledCache(RulesHandledCache):

	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_things_rulebooks_cache.retrieve(
				character, branch, turn, tick)
		except KeyError:
			return character, 'character_thing'

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			rulebook = self.get_rulebook(character, branch, turn, tick)
			things = sort_set(charm[character].thing.keys())
			pass
			for thing in things:
				try:
					rules = self.unhandled_rulebook_rules(
						character, thing, rulebook, branch, turn, tick)
				except KeyError:
					continue
				for rule in rules:
					yield character, thing, rulebook, rule


class CharacterPlaceRulesHandledCache(RulesHandledCache):

	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_places_rulebooks_cache.retrieve(
				character, branch, turn, tick)
		except KeyError:
			return character, 'character_place'

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			rulebook = self.get_rulebook(character, branch, turn, tick)
			for place in sort_set(charm[character].place.keys()):
				try:
					rules = self.unhandled_rulebook_rules(
						character, place, rulebook, branch, turn, tick)
				except KeyError:
					continue
				for rule in rules:
					yield character, place, rulebook, rule


class CharacterPortalRulesHandledCache(RulesHandledCache):

	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_portals_rulebooks_cache.retrieve(
				character, branch, turn, tick)
		except KeyError:
			return character, 'character_portal'

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			try:
				rulebook = self.get_rulebook(character, branch, turn, tick)
			except KeyError:
				continue
			char = charm[character]
			charn = char.node
			charp = char.portal
			for orig in sort_set(charp.keys()):
				if orig not in charn:
					continue
				for dest in sort_set(charp[orig].keys()):
					if dest not in charn:
						continue
					try:
						rules = self.unhandled_rulebook_rules(
							character, orig, dest, rulebook, branch, turn,
							tick)
					except KeyError:
						continue
					for rule in rules:
						yield character, orig, dest, rulebook, rule


class NodeRulesHandledCache(RulesHandledCache):

	def get_rulebook(self, character, node, branch, turn, tick):
		try:
			return self.engine._nodes_rulebooks_cache.retrieve(
				character, node, branch, turn, tick)
		except KeyError:
			return character, node

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			for node in sort_set(charm[character].node.keys()):
				try:
					rulebook = self.get_rulebook(character, node, branch, turn,
													tick)
					rules = self.unhandled_rulebook_rules(
						character, node, rulebook, branch, turn, tick)
				except KeyError:
					continue
				for rule in rules:
					yield character, node, rulebook, rule


class PortalRulesHandledCache(RulesHandledCache):

	def get_rulebook(self, character, orig, dest, branch, turn, tick):
		try:
			return self.engine._portals_rulebooks_cache.retrieve(
				character, orig, dest, branch, turn, tick)
		except KeyError:
			return character, orig, dest

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			char = charm[character]
			charn = char.node
			charp = char.portal
			for orig in sort_set(charp.keys()):
				if orig not in charn:
					continue
				dests = charp[orig]
				for dest in sort_set(dests.keys()):
					if dest not in charn:
						continue
					try:
						rulebook = self.get_rulebook(character, orig, dest,
														branch, turn, tick)
						rules = self.unhandled_rulebook_rules(
							character, orig, dest, rulebook, branch, turn,
							tick)
					except KeyError:
						continue
					for rule in rules:
						yield character, orig, dest, rulebook, rule


class ThingsCache(Cache):

	def __init__(self, db):
		Cache.__init__(self, db)
		self._make_node = db.thing_cls

	def store(self, *args, planning=None, loading=False, contra=True):
		character, thing, branch, turn, tick, location = args
		try:
			oldloc = self.retrieve(character, thing, branch, turn, tick)
		except KeyError:
			oldloc = None
		super().store(*args, planning=planning, loading=loading, contra=contra)
		node_contents_cache = self.db._node_contents_cache
		# Cache the contents of nodes
		if oldloc is not None:
			oldconts_orig = node_contents_cache.retrieve(
				character, oldloc, branch, turn, tick)
			newconts_orig = oldconts_orig.difference({thing})
			node_contents_cache.store(character,
										oldloc,
										branch,
										turn,
										tick,
										newconts_orig,
										contra=False,
										loading=True)
			future_location_data = node_contents_cache.settings[branch].future(
				turn)
			todo = []
			for trn in future_location_data:
				for tck in future_location_data[trn]:
					char, loca, contents = future_location_data[trn][tck]
					if char == character and loca == oldloc:
						todo.append((trn, tck))
			for trn, tck in todo:
				node_contents_cache.store(character,
											oldloc,
											branch,
											trn,
											tck,
											node_contents_cache.retrieve(
												character, oldloc, branch, trn,
												tck).difference({thing}),
											planning=False,
											contra=False,
											loading=True)
		if location is not None:
			try:
				oldconts_dest = node_contents_cache.retrieve(
					character, location, branch, turn, tick)
			except KeyError:
				oldconts_dest = frozenset()
			newconts_dest = oldconts_dest.union({thing})
			node_contents_cache.store(character,
										location,
										branch,
										turn,
										tick,
										newconts_dest,
										contra=False,
										loading=True)
			future_location_data = node_contents_cache.settings[branch].future(
				turn)
			todo = []
			for trn in future_location_data:
				for tck in sorted(future_location_data[trn]):
					char, loca, contents = future_location_data[trn][tck]
					if char == character and loca == location:
						todo.append((trn, tck))
			for trn, tck in todo:
				node_contents_cache.store(character,
											location,
											branch,
											trn,
											tck,
											node_contents_cache.retrieve(
												character, location, branch,
												trn, tck).union({thing}),
											planning=False,
											contra=False,
											loading=True)

	def turn_before(self, character, thing, branch, turn):
		try:
			self.retrieve(character, thing, branch, turn, 0)
		except KeyError:
			pass
		return self.keys[(character, )][thing][branch].rev_before(turn)

	def turn_after(self, character, thing, branch, turn):
		try:
			self.retrieve(character, thing, branch, turn, 0)
		except KeyError:
			pass
		return self.keys[(character, )][thing][branch].rev_after(turn)


class NodeContentsCache(Cache):

	def _iter_future_contradictions(self, entity, key, turns, branch, turn,
									tick, value):
		return self.db._things_cache._iter_future_contradictions(
			entity, key, turns, branch, turn, tick, value)

	def slow_iter_contents(self, character, place, branch, turn, tick):
		branch_now, turn_now, tick_now = self.db._btt()
		self.db.time = branch, turn
		self.db.tick = tick
		for thing in self.db.character[character].thing.values():
			if thing['location'] == place:
				yield thing.name
		self.db.time = branch_now, turn_now
		self.db.tick = tick_now

	def remove(self, branch, turn, tick):
		"""Delete data on or after this tick

        On the assumption that the future has been invalidated.

        """
		for parent, entitys in list(self.parents.items()):
			for entity, keys in list(entitys.items()):
				for key, branchs in list(keys.items()):
					if branch in branchs:
						branhc = branchs[branch]
						if turn in branhc:
							trun = branhc[turn]
							if tick in trun:
								del trun[tick]
							trun.truncate(tick)
							if not trun:
								del branhc[turn]
						branhc.truncate(turn)
						if not branhc:
							del branchs[branch]
					if not branchs:
						del keys[key]
				if not keys:
					del entitys[entity]
			if not entitys:
				del self.parents[parent]
		for branchkey, branches in list(self.branches.items()):
			if branch in branches:
				branhc = branches[branch]
				if turn in branhc:
					trun = branhc[turn]
					if tick in trun:
						del trun[tick]
					trun.truncate(tick)
					if not trun:
						del branhc[turn]
				branhc.truncate(turn)
				if not branhc:
					del branches[branch]
			if not branches:
				del self.branches[branchkey]
		for keykey, keys in list(self.keys.items()):
			for key, branchs in list(keys.items()):
				if branch in branchs:
					branhc = branchs[branch]
					if turn in branhc:
						trun = branhc[turn]
						if tick in trun:
							del trun[tick]
						trun.truncate(tick)
						if not trun:
							del branhc[turn]
					branhc.truncate(turn)
					if not branhc:
						del branches[branch]
				if not branchs:
					del keys[key]
			if not keys:
				del self.keys[keykey]
		sets = self.settings[branch]
		if turn in sets:
			setsturn = sets[turn]
			if tick in setsturn:
				del setsturn[tick]
			setsturn.truncate(tick)
			if not setsturn:
				del sets[turn]
		sets.truncate(turn)
		if not sets:
			del self.settings[branch]
		presets = self.presettings[branch]
		if turn in presets:
			presetsturn = presets[turn]
			if tick in presetsturn:
				del presetsturn[tick]
			presetsturn.truncate(tick)
			if not presetsturn:
				del presets[turn]
		presets.truncate(turn)
		if not presets:
			del self.presettings[branch]
		for entity, brnch in list(self.keycache):
			if brnch == branch:
				kc = self.keycache[entity, brnch]
				if turn in kc:
					kcturn = kc[turn]
					if tick in kcturn:
						del kcturn[tick]
					kcturn.truncate(tick)
					if not kcturn:
						del kc[turn]
				kc.truncate(turn)
				if not kc:
					del self.keycache[entity, brnch]
		self.shallowest = OrderedDict()

	def truncate_loc(self, character, location, branch, turn, tick):
		"""Remove future data about a particular location

        Return True if I deleted anything, False otherwise.

        """
		r = False
		branches_turns = self.branches[character, location][branch]
		branches_turns.truncate(turn)
		if turn in branches_turns:
			bttrn = branches_turns[turn]
			if bttrn.future(tick):
				bttrn.truncate(tick)
				r = True
		keyses = self.keys[character, location]
		for keysbranches in keyses.values():
			if branch not in keysbranches:
				continue
			keysbranch = keysbranches[branch]
			if keysbranch.future(turn):
				keysbranch.truncate(turn)
				r = True
			if turn in keysbranch:
				keysbranchturn = keysbranch[turn]
				if keysbranchturn.future(tick):
					keysbranchturn.truncate(tick)
					r = True
		if branch in self.settings:
			for sets in (self.settings, self.presettings):
				sets_branch = sets[branch]
				if turn in sets_branch:
					sets_turn = sets_branch[turn]
					for tic, setting in list(sets_turn.future(tick).items()):
						if setting[:2] == (character, location):
							del sets_turn[tic]
							r = True
					if not sets_turn:
						del sets_branch[turn]
						assert r, "Found an empty cache when I didn't delete anything"
				for trn, tics in list(sets_branch.future(turn).items()):
					for tic, setting in list(tics.future(tick).items()):
						if setting[:2] == (character, location):
							del tics[tic]
							r = True
					if not tics:
						del sets_branch[trn]
						assert r, "Found an empty cache when I didn't delete anything"
				if not sets_branch:
					del sets[branch]
					assert r, "Found an empty cache when I didn't delete anything"
		self.shallowest = OrderedDict()
		return r



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import networkx as nx
import pytest
from LiSE.exc import AmbiguousUserError


@pytest.fixture(scope="function")
def someplace(engy):
	yield engy.new_character('physical').new_place('someplace')


def test_contents(someplace):
	stuff = [someplace.new_thing(i) for i in range(10)]
	assert len(someplace.content) == len(stuff)
	assert len(someplace.contents()) == len(stuff)
	for i in range(10):
		assert i in someplace.content
		assert someplace.content[i] == stuff[i]
	for that in stuff:
		assert that in someplace.contents()
	for that in someplace.contents():
		assert that in stuff
	elsewhere = someplace.character.new_place('elsewhere')
	fust = elsewhere.new_thing(11)
	assert fust not in someplace.contents()
	assert 11 not in someplace.content
	with pytest.raises(KeyError):
		someplace.content[11]


def test_portal(someplace):
	assert not someplace.portal
	assert 'there' not in someplace.portal
	there = someplace.character.new_place('there')
	assert not there.preportal
	assert 'someplace' not in there.preportal
	someplace.character.new_portal('someplace', 'there')
	assert someplace.portal
	assert 'there' in someplace.portal
	assert 'there' not in someplace.preportal
	assert there.preportal
	assert 'someplace' in there.preportal
	assert 'someplace' not in there.portal
	someplace.character.remove_edge('someplace', 'there')
	assert 'there' not in someplace.portal
	assert 'someplace' not in there.preportal


def test_user(someplace):
	with pytest.raises(AmbiguousUserError):
		someplace.user.only
	someone = someplace.engine.new_character('someone')
	someone.add_unit(someplace)
	assert someplace.user.only is someone
	assert 'someone' in someplace.user
	assert someplace.user['someone'] is someone
	noone = someplace.engine.new_character('noone')
	assert 'noone' not in someplace.user
	noone.add_unit(someplace)
	with pytest.raises(AmbiguousUserError):
		someplace.user.only
	assert 'noone' in someplace.user
	assert someplace.user['noone'] is noone


def test_rulebook(someplace):
	assert someplace.rulebook.name == ('physical', 'someplace')
	someplace.rulebook = 'imadeitup'
	assert someplace.rulebook.name == 'imadeitup'


def test_deletion_after_keyframe(engy):
	phys = engy.new_character('physical', data=nx.grid_2d_graph(8, 8))
	del phys.place[5, 5]
	assert (5, 5) not in phys.place
	assert (5, 5) not in list(phys.place)
	while engy.turn < 20:
		engy.next_turn()
	assert (5, 5) not in phys.place
	assert (5, 5) not in list(phys.place)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import re
from functools import reduce
from collections import defaultdict
from LiSE.engine import Engine
import pytest
import os
import shutil
import tempfile

pytestmark = [pytest.mark.slow]


@pytest.fixture(scope='module')
def college24_premade():
	directory = tempfile.mkdtemp('.')
	shutil.unpack_archive(
		os.path.join(os.path.abspath(os.path.dirname(__file__)),
						'college24_premade.tar.xz'), directory)
	with Engine(directory) as eng:
		yield eng
	shutil.rmtree(directory)


def roommate_collisions(college24_premade):
	"""Test queries' ability to tell that all of the students that share
    rooms have been in the same place.

    """
	engine = college24_premade
	done = set()
	for chara in engine.character.values():
		if chara.name in done:
			continue
		match = re.match(r'dorm(\d)room(\d)student(\d)', chara.name)
		if not match:
			continue
		dorm, room, student = match.groups()
		other_student = '1' if student == '0' else '0'
		student = chara
		other_student = engine.character['dorm{}room{}student{}'.format(
			dorm, room, other_student)]

		same_loc_turns = list(
			engine.turns_when(
				student.unit.only.historical('location') ==
				other_student.unit.only.historical('location')))
		assert same_loc_turns, "{} and {} don't seem to share a room".format(
			student.name, other_student.name)
		assert len(
			same_loc_turns
		) >= 6, "{} and {} did not share their room for at least 6 turns".format(
			student.name, other_student.name)

		done.add(student.name)
		done.add(other_student.name)


def test_roomie_collisions_premade(college24_premade):
	roommate_collisions(college24_premade)


def sober_collisions(college24_premade):
	"""Students that are neither lazy nor drunkards should all have been
    in class together at least once.

    """
	engine = college24_premade
	students = [
		stu for stu in engine.character['student_body'].stat['characters']
		if not (stu.stat['drunkard'] or stu.stat['lazy'])
	]

	assert students

	def sameClasstime(stu0, stu1):
		assert list(
			engine.turns_when(
				stu0.unit.only.historical('location') == stu1.unit.only.
				historical('location') == engine.alias('classroom'))
		), """{stu0} seems not to have been in the classroom 
                at the same time as {stu1}.
                {stu0} was there at turns {turns0}
                {stu1} was there at turns {turns1}""".format(
			stu0=stu0.name,
			stu1=stu1.name,
			turns0=list(
				engine.turns_when(
					stu0.unit.only.historical('location') == engine.alias(
						'classroom'))),
			turns1=list(
				engine.turns_when(
					stu1.unit.only.historical('location') == engine.alias(
						'classroom'))))
		return stu1

	reduce(sameClasstime, students)


def test_sober_collisions_premade(college24_premade):
	sober_collisions(college24_premade)


def noncollision(college24_premade):
	"""Make sure students *not* from the same room never go there together"""
	engine = college24_premade
	dorm = defaultdict(lambda: defaultdict(dict))
	for character in engine.character.values():
		match = re.match(r'dorm(\d)room(\d)student(\d)', character.name)
		if not match:
			continue
		d, r, s = match.groups()
		dorm[d][r][s] = character
	for d in dorm:
		other_dorms = [dd for dd in dorm if dd != d]
		for r in dorm[d]:
			other_rooms = [rr for rr in dorm[d] if rr != r]
			for stu0 in dorm[d][r].values():
				for rr in other_rooms:
					for stu1 in dorm[d][rr].values():
						assert not list(
							engine.turns_when(
								stu0.unit.only.historical('location') ==
								stu1.unit.only.historical('location') ==
								engine.alias('dorm{}room{}'.format(d, r)))
						), "{} seems to share a room with {}".format(
							stu0.name, stu1.name)
				common = 'common{}'.format(d)
				for dd in other_dorms:
					for rr in dorm[dd]:
						for stu1 in dorm[dd][rr].values():
							assert not list(
								engine.turns_when(
									stu0.unit.only.historical('location') ==
									stu1.unit.only.historical(
										'location') == engine.alias(common))
							), "{} seems to have been in the same common room  as {}".format(
								stu0.name, stu1.name)


def test_noncollision_premade(college24_premade):
	noncollision(college24_premade)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector,  zacharyspector@gmail.com
"""Initialize the kobold test and launch the GUI."""
from ELiDE.app import ELiDEApp
from kobold import clear_off, mkengine, inittest

clear_off()
with mkengine() as engine:
	inittest(engine)

app = ELiDEApp()
app.run()
